from fractions import Fraction
from typing import Dict, Any, List

from DPGM import Node
from assignments import Assignment
import assignments as asg

DEFAULT_PRECISION = 5


class Distribution:

    def __init__(self, distribution: Dict[Assignment, Any], all_nodes: List[Node]):
        self.all_nodes = sorted(all_nodes)
        self.distribution = distribution
        self.z3_model = None
        self.model = None

    def get_assignment_probability(self, a: Assignment):
        return self.distribution[a]

    def get_marginal_probability(self, node: Node):
        probability = 0
        for assignment in asg.get_node_assignments(node, self.all_nodes):
            probability += self.get_assignment_probability(assignment)
        return probability

    def print_assignment_prob(self, nodes=None, only_positive=False, printer=print, precision=DEFAULT_PRECISION):
        """
        Print distribution over the assignments of the given nodes or all nodes if no list is given.
        If the only_positive flag is set, only the assignments with probability greater than zero are outputted.
        """
        if not nodes:
            nodes = self.all_nodes
            node_assignments = asg.generate(nodes)
        else:
            node_assignments = self.distribution.keys()
        for assignment in node_assignments:
            value = self.get_assignment_probability(assignment)
            if not only_positive or round(value, precision + 1) > 0:
                print_assignment_probability(assignment, value, printer=printer, precision=precision)

    def print_formatted(self, format_string: str, printer=print, precision=DEFAULT_PRECISION):
        """
        Print distribution given by in the formats specified by the format string: (M) marginal node probabilities,
        (F) full distribution, (S) support of distribution, (Z) Z3 output, (T) SMT2 format.
        """
        for char in format_string:
            if char == "F":
                self.print_assignment_prob(printer=printer, precision=precision)
            elif char == "S":
                printer("Support:")
                self.print_assignment_prob(only_positive=True, printer=printer, precision=precision)
            elif char == "M":
                for node in self.all_nodes:
                    prob = self.get_marginal_probability(node)

                    printer("P(" + str(node.name) + ") = " + format_probability(prob, precision))
            elif char == "Z":
                if self.z3_model is not None:
                    printer(str(self.z3_model))
                elif self.model is not None:
                    printer(str(self.model))
                else:
                    printer("No Z3 model available.")
                    self.print_assignment_prob(printer=printer, precision=precision)
            elif char == "T":
                if self.z3_model is not None:
                    printer(str(self.z3_model.sexpr()))
                else:
                    printer("No Z3 model available, thus no SMT2 format as well.")
                    self.print_assignment_prob(printer=printer, precision=precision)
            else:
                printer("Unknown distribution format '" + char + "'.")
                self.print_assignment_prob(printer=printer, precision=precision)

    def __eq__(self, other):
        if isinstance(other, Distribution):
            for a in self.distribution.keys():
                if self.get_assignment_probability(a) != other.get_assignment_probability(a):
                    return False
            return True
        return False


def format_probability(prob: float, precision=DEFAULT_PRECISION):
    if isinstance(prob, Fraction):
        return str(prob)
    return "{:0.{p}g}".format(round(prob, precision+1), p=precision)


def repr_assignment_prob(a: Assignment, prob: float, given: Assignment = None, precision=DEFAULT_PRECISION):
    cond = "P(" + asg.repr_assignment(a) + ("|" + asg.repr_assignment(given) if given else "") + ")"
    return cond + " = " + format_probability(prob, precision)


def print_assignment_probability(a: Assignment, prob: float, precision=DEFAULT_PRECISION, printer=print):
    printer(repr_assignment_prob(a, prob, None, precision=precision))



from typing import List, Set

import z3
import pysmt.shortcuts as smt

from DPGM import Node
from distribution import Distribution
from labeling import Labeling
from labeling_scheme import LabelingScheme
from linear_solver import LinearSolver, LinearConstraintError
from semantics import Semantics
from solver_instance import SolverInstance
from timer import Timer
from z3_instance import Z3Instance


def ls_constraints(ls: LinearSolver, semantics: List[Semantics]):
    semantics_constraints_timer = Timer(output_string="Generating constraints for all semantics took")
    constraints = []
    for sem in semantics:
        try:
            constraints.extend(sem.get_linear_constraints(ls))
        except NotImplementedError:
            raise LinearConstraintError("No linear constraints available (or implemented) for semantics '"
                                        + str(sem) + "'") from None
    semantics_constraints_timer.stop()
    return constraints


def get_model_z3(z3i: Z3Instance, semantics: List[Semantics], constraints=None):
    """
    Check if a distribution satisfying the constraints of all semantics plus potential additional constraints exists.
    If so, return it as a model.

    :param z3i: the underlying z3 instance with the AF
    :param semantics: a list of semantics
    :param constraints: optional list of additional constraints
    :return: A model if one exists, None otherwise
    """
    if not constraints:
        constraints = []

    semantics_constraints_timer = Timer(output_string="Generating constraints for all semantics took")
    for sem in semantics:
        # print(sem.__class__.__name__, sem.get_z3_constraints(z3i))
        constraints.extend(sem.get_z3_constraints(z3i))
    semantics_constraints_timer.stop()

    result = z3i.check_satisfiability(constraints)
    if result == z3.sat:
        return z3i.solver.model()
    else:
        return None


def get_model(si: SolverInstance, semantics: List[Semantics], constraints=None):
    """
    Check if a distribution satisfying the constraints of all semantics plus potential additional constraints exists.
    If so, return it as a model.

    :param si: the interface to the smt solvers
    :param semantics: a list of semantics
    :param constraints: optional list of additional constraints
    :return: A model if one exists, None otherwise
    """
    if not constraints:
        constraints = []

    semantics_constraints_timer = Timer(output_string="Generating constraints for all semantics took")
    for sem in semantics:
        # print(str(sem) + ":", sem.get_constraints(si))
        constraints.extend(sem.get_constraints(si))
    semantics_constraints_timer.stop()

    return si.get_model(constraints)


def get_one_distribution(si: SolverInstance, semantics: List[Semantics]) -> Distribution:
    """
    Return one distribution satisfying all constraints if one exists.

    :param si: the interface to the smt solvers
    :param semantics: a list of semantics
    :return: one distribution if one exists
    """
    model = get_model(si, semantics)
    if model:
        return si.distribution_from_model(model)


def get_one_distribution_z3(z3i: Z3Instance, semantics: List[Semantics]) -> Distribution:
    """
    Return one distribution satisfying all constraints if one exists.

    :param z3i: the underlying z3 instance with the AF
    :param semantics: a list of semantics
    :return: one distribution if one exists
    """
    model = get_model_z3(z3i, semantics)
    if model:
        return z3i.distribution_from_model(model)


def get_one_distribution_linear(ls: LinearSolver, semantics: List[Semantics]) -> Distribution:
    """
    Return one distribution satisfying all constraints if one exists.
    Semantics that do not admit linear constraints are skipped.

    :param ls: the linear solver instance
    :param semantics: a list of semantics
    :return: one distribution if one exists
    """
    constraints = ls_constraints(ls, semantics)
    return ls.any_solution(constraints)


def get_corner_distributions(ls: LinearSolver, semantics: List[Semantics]) -> List[Distribution]:
    """
    Compute the corner distribution of the convex hull of the solution space using a linear solver. The infinite set of
    all solutions is then given by the convex combinations of these distributions.

    Semantics that do not admit linear constraints are skipped.

    :param ls: the linear solver instance
    :param semantics: a list of semantics
    :return: the distributions forming corners of the convex solution space
    """
    constraints = ls_constraints(ls, semantics)
    return ls.corner_distributions(constraints)


def get_satisfying_labeling_z3(z3i: Z3Instance, lab: LabelingScheme, semantics: List[Semantics]) -> Labeling:
    """
    Get a labeling satisfying the given semantics under the given labeling scheme. Returns None if no such labeling
    exists.

    :param z3i: the underlying z3 instance with the AF
    :param lab: a labeling scheme
    :param semantics: a list of semantics
    :return: A labeling or None
    """
    model = get_model_z3(z3i, semantics)
    if model:
        return Labeling(lab, z3i.distribution_from_model(model), z3i.af)


def get_all_satisfying_labelings(si: SolverInstance, lab: LabelingScheme, semantics: List[Semantics]) -> Set[Labeling]:
    """
    Compute all labelings under the given labeling scheme that satisfy the given semantics.

    :param si: the interface to the smt solvers
    :param lab: a labeling scheme
    :param semantics: a list of semantics
    :return: a list of labelings
    """
    semantics_constraints_timer = Timer(output_string="Generating constraints for all semantics took")
    constraints = []
    for sem in semantics:
        constraints.extend(sem.get_constraints(si))
    semantics_constraints_timer.stop()

    solver = si.get_solver()
    solver.add_assertion(smt.And(constraints))

    # test overall satisfiability
    sat = solver.solve()
    # print("Overall satisfiability:", sat)
    if not sat:
        return set()

    candidate_constraints = [[]]
    labelings = set()
    num_of_checks = 1
    num_remaining_nodes = len(si.af.get_nodes())
    for node in si.af.get_nodes():
        num_remaining_nodes -= 1
        # print(node.name)
        next_candidate_constraints = []
        for candidate_constraint in candidate_constraints:
            for const in lab.get_constraints(node, si):
                new_candidate_constraint = candidate_constraint.copy()
                new_candidate_constraint.append(const)

                result = solver.is_sat(smt.And(new_candidate_constraint))
                num_of_checks += 1
                if result:
                    # print("new_candidate_constraint", new_candidate_constraint)
                    next_candidate_constraints.append(new_candidate_constraint)
                    # add labelings if we are at the leaves of the search tree
                    if num_remaining_nodes == 0:
                        model = solver.get_model()
                        distribution = si.distribution_from_model(model)
                        labeling = Labeling(lab, distribution, si.af)
                        # assert labeling not in labelings, "Labeling found twice: " + str(labeling)
                        labelings.add(labeling)
        candidate_constraints = next_candidate_constraints

    # How many constraints do we have for each node? Usually as many as there are labelings, though for some labeling
    # schemes there are fewer
    branching = len(lab.get_constraints(list(si.af.get_nodes())[0], si))
    # number of nodes in a perfect k-ary tree of depth n = number of nodes and k = branching,
    # i.e., wc = (k**(n+1) - 1) // (k-1)
    worst_case_checks = (branching**(len(si.af.get_nodes()) + 1) - 1) // (branching - 1)

    print("Finished. Performed {} checks (worst case: {}) and found {} labelings.".format(
        num_of_checks, worst_case_checks, len(labelings)))
    return labelings


def get_all_satisfying_labelings_z3(z3i: Z3Instance, lab: LabelingScheme, semantics: List[Semantics]) -> Set[Labeling]:
    """
    Compute all labelings under the given labeling scheme that satisfy the given semantics.

    :param z3i: the underlying z3 instance with the AF
    :param lab: a labeling scheme
    :param semantics: a list of semantics
    :return: a list of labelings
    """
    semantics_constraints_timer = Timer(output_string="Generating constraints for all semantics took")
    semantics_constraints = []
    for sem in semantics:
        # print(sem.__class__.__name__, sem.get_z3_constraints(z3i))
        semantics_constraints.extend(sem.get_z3_constraints(z3i))
    semantics_constraints_timer.stop()

    z3i.solver.reset()
    z3i.solver.add(z3i.constraints)
    z3i.solver.add(z3i.edge_constraints)
    z3i.solver.add(semantics_constraints)

    # test overall satisfiability
    result = z3i.solver.check()
    # print("Overall satisfiability:", result)
    if result != z3.sat:
        return set()

    candidate_constraints = [[]]
    labelings = set()
    num_of_checks = 1
    num_remaining_nodes = len(z3i.af.get_nodes())

    for node in z3i.af.get_nodes():
        num_remaining_nodes -= 1
        # print(node.name)
        next_candidate_constraints = []
        for candidate_constraint in candidate_constraints:
            for const in lab.get_constraints_z3(node, z3i):
                new_candidate_constraint = candidate_constraint.copy()
                new_candidate_constraint.append(const)
                # add candidate constraints to solver
                z3i.solver.push()  # this adds an backtracking point
                # print(z3i.solver)
                z3i.solver.add(new_candidate_constraint)
                # print(z3i.solver)
                result = z3i.solver.check()
                num_of_checks += 1
                if result == z3.sat:
                    # print("new_candidate_constraint", new_candidate_constraint)
                    next_candidate_constraints.append(new_candidate_constraint)
                    # add labelings if we are at the leaves of the search tree
                    if num_remaining_nodes == 0:
                        model = z3i.solver.model()
                        labeling = Labeling(lab, z3i.distribution_from_model(model), z3i.af)
                        # assert labeling not in labelings, "Labeling found twice: " + str(labeling)
                        labelings.add(labeling)
                #  remove candidate constraints from solver
                z3i.solver.pop()  # go back to the backtracking point
        candidate_constraints = next_candidate_constraints

    # How many constraints do we have for each node? Usually as many as there are labelings, though for some labeling
    # schemes there are fewer
    branching = len(lab.get_constraints_z3(list(z3i.af.get_nodes())[0], z3i))
    # number of nodes in a perfect k-ary tree of depth n = number of nodes and k = branching,
    # i.e., wc = (k**(n+1) - 1) // (k-1)
    worst_case_checks = (branching**(len(z3i.af.get_nodes()) + 1) - 1) // (branching - 1)

    print("Finished. Performed {} checks (worst case: {}) and found {} labelings.".format(
        num_of_checks, worst_case_checks, len(labelings)))
    return labelings


def compare_semantics_by_labelings_z3(z3i: Z3Instance, lab: LabelingScheme, semantics_a: List[Semantics],
                                      semantics_b: List[Semantics]):
    """

    :param z3i: the underlying z3 instance
    :param lab: the labeling scheme
    :param semantics_a: the first set of semantics
    :param semantics_b: the second set of semantics
    :return: three sets of labelings for `A u B`, `A - B` and `B - A`
    """

    labelings_a = set(get_all_satisfying_labelings_z3(z3i, lab, semantics_a))
    labelings_b = set(get_all_satisfying_labelings_z3(z3i, lab, semantics_b))

    labelings_both = labelings_a.union(labelings_b)
    labelings_only_a = labelings_a.difference(labelings_b)
    labelings_only_b = labelings_b.difference(labelings_a)
    return labelings_both, labelings_only_a, labelings_only_b


def check_credulous_acceptance(si: SolverInstance, semantics: List[Semantics], argument: Node, threshold: float = 1):
    """
    Check if the given argument is credulously accepted, i.e. there exists at least one distribution satisfying the
    given semantics that assigns the argument a probability greater or equal to the threshold (equal to 1 by default).

    :param si: the interface to the smt solvers
    :param semantics: a list of semantics
    :param argument: an argument from the AF
    :param threshold: optional: a float from the interval [0,1]
    :return: True + a model if one exists, False otherwise
    """
    argument_symbol = si.get_node_symbol(argument)
    constraints = [smt.GE(argument_symbol, smt.Real(threshold))]
    model = get_model(si, semantics, constraints)
    if model:
        distribution = si.distribution_from_model(model)
        return True, distribution
    else:
        return False, None


def check_credulous_acceptance_z3(z3i: Z3Instance, semantics: List[Semantics], argument: Node, threshold: float = 1):
    """
    Check if the given argument is credulously accepted, i.e. there exists at least one distribution satisfying the
    given semantics that assigns the argument a probability greater or equal to the threshold (equal to 1 by default).

    :param z3i: the underlying z3 instance with the AF
    :param semantics: a list of semantics
    :param argument: an argument from the AF
    :param threshold: optional: a float from the interval [0,1]
    :return: True + a model if one exists, False otherwise
    """
    argument_prob = z3i.get_node_prob_var(argument)
    constraints = [argument_prob >= threshold]
    model = get_model_z3(z3i, semantics, constraints)
    if model:
        distribution = z3i.distribution_from_model(model)
        return True, distribution
    else:
        return False, None


def check_skeptical_acceptance(si: SolverInstance, semantics: List[Semantics], argument: Node, threshold: float = 1):
    """
    Check if the given argument is skeptically accepted, i.e. all distributions that satisfy the given semantics assign
    the argument a probability greater or equal to the threshold (equal to 1 by default).

    :param si: the interface to the smt solvers
    :param semantics: a list of semantics
    :param argument: an argument from the AF
    :param threshold: optional: a float from the interval [0,1]
    :return: True or False + counter example distribution
    """
    # "check if for all satisfying distributions, P(A) >= t holds"
    # <=> not "check if for one satisfying distribution, P(A) < t holds"
    argument_symbol = si.get_node_symbol(argument)
    constraints = [smt.LT(argument_symbol, smt.Real(threshold))]
    model = get_model(si, semantics, constraints)
    if model:
        distribution = si.distribution_from_model(model)
        return False, distribution
    else:
        return True, None


def check_skeptical_acceptance_z3(z3i: Z3Instance, semantics: List[Semantics], argument: Node, threshold: float = 1):
    """
    Check if the given argument is skeptically accepted, i.e. all distributions that satisfy the given semantics assign
    the argument a probability greater or equal to the threshold (equal to 1 by default).

    :param z3i: the underlying z3 instance with the AF
    :param semantics: a list of semantics
    :param argument: an argument from the AF
    :param threshold: optional: a float from the interval [0,1]
    :return: True or False + counter example distribution
    """
    # "check if for all satisfying distributions, P(A) >= t holds"
    # <=> not "check if for one satisfying distribution, P(A) < t holds"
    argument_prob = z3i.get_node_prob_var(argument)
    constraints = [argument_prob < threshold]
    model = get_model_z3(z3i, semantics, constraints)
    if model:
        distribution = z3i.distribution_from_model(model)
        return False, distribution
    else:
        return True, None


def maximize_marginal_probabilities(ls: LinearSolver, nodes: List[Node], semantics: List[Semantics]):
    """
    Compute the optimal distribution which maximizes the marginal probability of the given arguments while observing
    the given semantics' constraints.
    """
    constraints = ls_constraints(ls, semantics)
    return ls.optimize_marginal_prob(nodes, constraints, mode="max")


def minimize_marginal_probabilities(ls: LinearSolver, nodes: List[Node], semantics: List[Semantics]):
    """
    Compute the optimal distribution which minimizes the marginal probability of the given arguments while observing
    the given semantics' constraints.
    """
    constraints = ls_constraints(ls, semantics)
    return ls.optimize_marginal_prob(nodes, constraints, mode="min")



import argparse
from typing import List

import af_parser
import labeling_scheme as lab
from labeling import Labeling
from linear_solver import LinearSolver, LinearConstraintError
from DPGM import DPGM
import semantics as sem
import tasks
from solver_instance import SolverInstance
from z3_instance import Z3Instance
from timer import Timer

# TODO:
#  - produce log files
#  - task to get distribution unequal to a given other distribution?
#  - re-add assume-independence flag
#
#  - Flag for extensions
#  - flag to generate unsat core; this needs named assertions, see assert_and_track() in z3.py
#  - credulous and skeptical acceptance wrt. a labeling scheme

__version__ = "0.7.0"

# Solvers
Z3_LEGACY = "z3-legacy"
# from https://pysmt.readthedocs.io/en/latest/getting_started.html#the-basics
# noinspection SpellCheckingInspection
pysmt_solvers = ["z3", "msat", "cvc4", "yices", "btor", "picosat", "bdd"]
linear_solvers = ["cvxopt", "glpk", "mosek"]
other_solvers = [Z3_LEGACY]
all_solvers = pysmt_solvers + linear_solvers + other_solvers


def main():
    main_timer = Timer(output_string="Overall runtime:")

    #########################
    # set up argument parser
    #########################
    arg_parser = argparse.ArgumentParser(prog="cpraa", add_help=False,
                                         description="CPrAA - a Checker for Probabilistic Abstract Argumentation")
    base_group = arg_parser.add_argument_group("basic parameters")
    base_group.add_argument("-f", "--file", help="the argumentation framework as .tgf file")
    base_group.add_argument("-s", "--semantics", nargs="+", default=[],
                            help="the probabilistic semantics to check against")
    base_group.add_argument("-cs", "--complement_semantics", nargs="+", default=[], metavar="SEMANTICS",
                            help="probabilistic semantics with negated constraints")
    base_group.add_argument("-a", "--arguments", nargs="+", default=[], metavar="NODE",
                            help="one or more nodes from the AF to be used in one of tasks")
    base_group.add_argument("-OD", "--one_distribution", action="store_true",
                            help="compute one distribution satisfying the semantics")
    base_group.add_argument("-CD", "--corner_distributions", action="store_true",
                            help="compute the distributions forming the corners of the convex hull of the solution "
                                 "space spanned by the semantics' linear constraints")
    # base_group.add_argument("-i", "--assume_independence", action="store_true",
    #                         help="treat the arguments' probabilities as independent "
    #                              "(warning: this leads to incorrect results for some semantics)")
    base_group.add_argument("-b", "--backend_solver", metavar="SOLVER", type=str,
                            help="the solver used as backend. Options are all SMT solvers available via pySMT ("
                                 + ", ".join(pysmt_solvers) + "), linear solvers via CVXOPT ("
                                 + ", ".join(linear_solvers) + "), and a direct integration of Z3 ("
                                 + Z3_LEGACY + ")")
    base_group.add_argument("-smt", "--smt_file", help="add additional constraints from a file in SMT-LIB format")

    lp_group = arg_parser.add_argument_group("optimization")
    lp_group.add_argument("-MIN", "--minimize_probability", action="store_true",
                          help="find the distribution minimizing the marginal probabilities of the arguments passed "
                               "with -a while satisfying the given semantics")
    lp_group.add_argument("-MAX", "--maximize_probability", action="store_true",
                          help="find the distribution maximizing the marginal probabilities of the arguments passed "
                               "with -a while satisfying the given semantics")

    ac_group = arg_parser.add_argument_group("acceptance checking")
    ac_group.add_argument("-CA", "--credulous_acceptance",
                          nargs="?", metavar="THRESHOLD", type=float, const=1.0,
                          help="check credulous acceptance of given arguments, "
                               "optionally taking a threshold into account")
    ac_group.add_argument("-SA", "--skeptical_acceptance",
                          nargs="?", metavar="THRESHOLD", type=float, const=1.0,
                          help="check skeptical acceptance of given arguments, "
                               "optionally taking a threshold into account")
    ac_group.add_argument("-CAA", "--credulous_acceptance_all",
                          nargs="?", metavar="THRESHOLD", type=float, const=1.0,
                          help="check credulous acceptance of all arguments, "
                               "optionally taking a threshold into account")
    ac_group.add_argument("-SAA", "--skeptical_acceptance_all",
                          nargs="?", metavar="THRESHOLD", type=float, const=1.0,
                          help="check skeptical acceptance of all arguments, "
                               "optionally taking a threshold into account")

    lab_group = arg_parser.add_argument_group("generate labelings")
    lab_group.add_argument("-l", "--labeling_scheme", metavar="SCHEME", help="the labeling scheme to use")
    lab_group.add_argument("-OL", "--one_labeling", action="store_true",
                           help="compute one labeling satisfying the semantics")
    lab_group.add_argument("-AL", "--all_labelings", action="store_true",
                           help="compute all labelings satisfying the semantics")
    lab_group.add_argument("-lt", "--labeling_threshold", nargs="+", metavar="T", type=float,
                           help="the optional thresholds needed for some labeling schemes")

    info_group = arg_parser.add_argument_group('informational parameters')
    info_group.add_argument("-h", "--help", action="help", help="show this help message")
    info_group.add_argument("-v", "--version", action="version", version="%(prog)s " + __version__)
    info_group.add_argument("-ls", "--list_semantics", action="store_true",
                            help="list all available probabilistic semantics")
    info_group.add_argument("-ll", "--list_labeling_schemes", action="store_true",
                            help="list all available labeling schemes")
    info_group.add_argument("-d", "--documentation", nargs="+", default=[], metavar="NAME",
                            help="display a short documentation for the listed semantics or labeling schemes")
    info_group.add_argument("-t", "--time", action="store_true", help="time the execution")
    info_group.add_argument("-o", "--distribution_output_format", metavar="FORMAT", type=str, default="S",
                            help="how to output distributions: (M) marginal node probabilities, (F) full distribution, "
                                 "(S) support of distribution, (Z) Z3 output, (T) SMT2 format.")
    info_group.add_argument("-p", "--output_precision", metavar="PRECISION", type=int, default=5,
                            help="precision of floats that are output")
    info_group.add_argument("-dc", "--display_constraints", action="store_true",
                            help="print the constraints generated by the semantics (note that for complemented "
                                 "semantics, the constraints are displayed without the subsequent conjunction and "
                                 "negation)")

    ###############################
    # parse command line arguments
    ###############################
    args = arg_parser.parse_args()

    af = None
    solver = None
    z3i = None
    ls = None
    si = None
    labeling_scheme = None
    semantics = []
    arguments = None
    args.assume_independence = False  # legacy option

    main_timer.enable_printing = args.time
    Timer.enable_printing_default = args.time

    sem.CONSTRAINT_LOGGING = args.display_constraints

    if args.list_semantics:
        print("Available semantics:", ", ".join(reversed(sem.all_semantics_names())))

    if args.list_labeling_schemes:
        print("Available labeling schemes:", ", ".join(reversed(lab.get_all_labeling_scheme_names())))

    if args.documentation:
        for name in args.documentation:
            semantics_class = sem.get_semantics_class_by_name(name)
            if semantics_class:
                print("Semantics '" + semantics_class.__name__ + "':")
                if semantics_class.__doc__:
                    print(semantics_class.__doc__.strip("\n"))
                else:
                    print("    No documentation available.")
            else:
                labeling_scheme_class = lab.get_labeling_scheme_class(name)
                if labeling_scheme_class:
                    print("Labeling scheme '" + labeling_scheme_class.__name__ + "':")
                    if labeling_scheme_class.__doc__:
                        print(labeling_scheme_class.__doc__.strip("\n"))
                    else:
                        print("    No documentation available.")
                else:
                    arg_parser.error("no semantics or labeling scheme called '" + name + "' exists")

    if args.file:
        af = af_parser.parse_tgf(args.file)
        # print(af)

    if args.backend_solver:
        if not af:
            arg_parser.error("no AF provided")
        solver = args.backend_solver
        if solver not in all_solvers:
            arg_parser.error("Unknown solver name '" + solver + "'. Possible solvers are " + ", ".join(all_solvers)
                             + " (though some of those need to be installed separately)")

    if args.semantics:
        if not af:
            arg_parser.error("an AF must be provided when checking semantics")

        for semantics_name in args.semantics:
            semantics_class = sem.get_semantics_class_by_name(semantics_name)
            if not semantics_class:
                arg_parser.error("No semantics called '"
                                 + semantics_name + "' exists. Use -ls to list all available semantics.")
            semantics.append(semantics_class(af))

    if args.complement_semantics:
        if not af:
            arg_parser.error("an AF must be provided when checking semantics")

        for semantics_name in args.complement_semantics:
            semantics_class = sem.get_semantics_class_by_name(semantics_name)
            if not semantics_class:
                arg_parser.error("No semantics called '"
                                 + semantics_name + "' exists. Use -ls to list all available semantics.")
            semantics.append(semantics_class(af, complemented=True))

    if args.labeling_scheme:
        labeling_scheme_class = lab.get_labeling_scheme_class(args.labeling_scheme)
        if not labeling_scheme_class:
            arg_parser.error("no labeling scheme called '" + args.labeling_scheme + "' exists")
        num_args_required = labeling_scheme_class.num_args
        if args.labeling_threshold:
            num_args_given = len(args.labeling_threshold)
        else:
            num_args_given = 0
        if num_args_given < num_args_required:
            error_message = "mismatch in the number of thresholds: required by labeling scheme '" \
                            + args.labeling_scheme + "': " + str(num_args_required) + ", given: " + str(num_args_given)
            arg_parser.error(error_message)
        if num_args_given > num_args_required:
            warning_message = "warning: mismatch in the number of thresholds: required by labeling scheme '" \
                              + args.labeling_scheme + "': " + str(num_args_required) + ", given: " \
                              + str(num_args_given) + " - i'm ignoring the surplus ones"
            print(warning_message)
        if num_args_required == 0:
            labeling_scheme = labeling_scheme_class()
        elif num_args_required == 1:
            labeling_scheme = labeling_scheme_class(args.labeling_threshold[0])
        elif num_args_required == 2:
            labeling_scheme = labeling_scheme_class(args.labeling_threshold[0], args.labeling_threshold[1])
        else:
            raise NotImplementedError("labeling schemes with more than 2 thresholds")

    if args.arguments:
        if not af:
            arg_parser.error("no AF provided")
        try:
            arguments = [af.get_node_by_name(argument_name) for argument_name in args.arguments]
        except ValueError as e:
            arg_parser.error(str(e))

    if args.smt_file:
        if not af:
            arg_parser.error("no AF provided")
        if solver == Z3_LEGACY or solver in linear_solvers:
            arg_parser.error("smt constraints only available via pysmt solvers")
        # solver in pysmt_solvers or solver is None
        si = get_smt_solver(si, af, solver)
        with open(args.smt_file, "r") as smt_file:
            si.add_constraints_from_smt(smt_file)

    ########
    # tasks
    ########

    if args.one_distribution:
        if not af:
            arg_parser.error("an AF must be provided to compute a distribution")

        od_timer = Timer(output_string="Computing one distribution took", start=False)
        print("Computing one distribution satisfying the following semantics:", repr_list(semantics))

        if solver == Z3_LEGACY:
            z3i = get_z3_solver(z3i, af, args.assume_independence)
            od_timer.start()
            distribution = tasks.get_one_distribution_z3(z3i, semantics)
        elif solver in linear_solvers:
            ls = get_linear_solver(ls, af, solver)
            od_timer.start()
            distribution = tasks.get_one_distribution_linear(ls, semantics)
        else:  # solver in pysmt_solvers or solver is None
            si = get_smt_solver(si, af, solver)
            od_timer.start()
            distribution = tasks.get_one_distribution(si, semantics)

        if distribution:
            distribution.print_formatted(args.distribution_output_format, precision=args.output_precision)
        else:
            print("No distribution satisfying the given semantics exists.")
        od_timer.stop()

    if args.corner_distributions:
        if not af:
            arg_parser.error("an AF must be provided to compute the corner distribution")
        if args.smt_file:
            arg_parser.error("task not available with additional SMT constraints")
        assert_linear_semantics_availability(semantics, arg_parser)
        if solver in linear_solvers:
            ls = get_linear_solver(ls, af, solver)
        else:
            ls = get_linear_solver(ls, af, None)

        cd_timer = Timer(output_string="Computing the corner distributions took")
        print("Computing the corner distributions for the following semantics:", repr_list(semantics))
        distributions = tasks.get_corner_distributions(ls, semantics)
        num_results = len(distributions)
        if num_results == 0:
            print("No distribution satisfying the given semantics exists.")
        elif num_results == 1:
            print("The solution is unique:")
            distributions[0].print_formatted(args.distribution_output_format)
        else:
            for i in range(num_results):
                print("\nResult {} of {}:".format(i + 1, num_results))
                distributions[i].print_formatted(args.distribution_output_format)
        cd_timer.stop()

    if args.minimize_probability:
        if not af:
            arg_parser.error("no AF given, use -f to load one from a TGF file")
        if not arguments:
            arg_parser.error("no arguments from the AF given, provide one or more with -a")
        if args.smt_file:
            arg_parser.error("task not available with additional SMT constraints")
        assert_linear_semantics_availability(semantics, arg_parser)
        if solver in linear_solvers:
            ls = get_linear_solver(ls, af, solver)
        else:
            ls = get_linear_solver(ls, af, None)
        timer = Timer(output_string="Computing optimal distribution took")
        print("Computing optimal distribution minimizing the probability of argument" + plural(args.arguments),
              repr_list_i(args.arguments), "while satisfying the following semantics:", repr_list(semantics))
        distribution = tasks.minimize_marginal_probabilities(ls, arguments, semantics)
        if distribution:
            distribution.print_formatted(args.distribution_output_format)
        else:
            print("No distribution satisfying the given semantics exists.")
        timer.stop()

    if args.maximize_probability:
        if not af:
            arg_parser.error("no AF given, use -f to load one from a TGF file")
        if not arguments:
            arg_parser.error("no arguments from the AF given, provide one or more with -a")
        if args.smt_file:
            arg_parser.error("task not available with additional SMT constraints")
        assert_linear_semantics_availability(semantics, arg_parser)
        if solver in linear_solvers:
            ls = get_linear_solver(ls, af, solver)
        else:
            ls = get_linear_solver(ls, af, None)
        timer = Timer(output_string="Computing optimal distribution took")
        print("Computing optimal distribution maximizing the probability of argument" + plural(args.arguments),
              repr_list_i(args.arguments), "while satisfying the following semantics:", repr_list(semantics))
        distribution = tasks.maximize_marginal_probabilities(ls, arguments, semantics)
        if distribution:
            distribution.print_formatted(args.distribution_output_format)
        else:
            print("No distribution satisfying the given semantics exists.")
        timer.stop()

    if args.one_labeling:
        if not af:
            arg_parser.error("an AF must be provided to compute a labeling")
        if not labeling_scheme:
            arg_parser.error("no labeling scheme provided")

        ol_timer = Timer(output_string="Computing one labeling took", start=False)
        print("Computing one", args.labeling_scheme, "labeling satisfying the following semantics:",
              repr_list(semantics))

        if solver == Z3_LEGACY:
            z3i = get_z3_solver(z3i, af, args.assume_independence)
            ol_timer.start()
            distribution = tasks.get_one_distribution_z3(z3i, semantics)
        elif solver in linear_solvers:
            assert_linear_semantics_availability(semantics, arg_parser)
            ls = get_linear_solver(ls, af, solver)
            ol_timer.start()
            distribution = tasks.get_one_distribution_linear(ls, semantics)
        else:  # solver in pysmt_solvers or solver is None
            si = get_smt_solver(si, af, solver)
            ol_timer.start()
            distribution = tasks.get_one_distribution(si, semantics)

        if distribution:
            labeling = Labeling(labeling_scheme, distribution, af)
            print(labeling)
        else:
            print("No distribution satisfying the given semantics exists.")
        ol_timer.stop()

    if args.all_labelings:
        if not af:
            arg_parser.error("an AF must be provided to compute a labeling")
        if not labeling_scheme:
            arg_parser.error("no labeling scheme provided (use -l)")

        if solver in linear_solvers:
            arg_parser.error("all labelings task not available with linear solver '" + solver + "'")
        elif solver == Z3_LEGACY:
            z3i = get_z3_solver(z3i, af, args.assume_independence)
        else:  # solver in pysmt_solvers or solver is None
            si = get_smt_solver(si, af, solver)

        al_timer = Timer(output_string="Computing all labelings took")
        print("Computing all", args.labeling_scheme, "labelings satisfying the following semantics:",
              repr_list(semantics))
        if solver == Z3_LEGACY:
            labelings = tasks.get_all_satisfying_labelings_z3(z3i, labeling_scheme, semantics)
        else:
            labelings = tasks.get_all_satisfying_labelings(si, labeling_scheme, semantics)
        print("Number of labelings:", len(labelings))
        # for ext in sorted([labeling.labeled_in for labeling in labelings]):
        #     print(ext)
        for labeling in sorted(labelings):
            print(labeling)
        # for labeling in sorted(list(set(map(repr, labelings)))):
        #     print(labeling)
        al_timer.stop()

    if args.credulous_acceptance is not None:
        if not af:
            arg_parser.error("an AF must be provided to check acceptance")
        if not arguments:
            arg_parser.error("no arguments given to check acceptance, provide one or more with -a")

        if solver in linear_solvers:
            arg_parser.error("credulous acceptance checking not available with linear solver '" + solver + "'")
        elif solver == Z3_LEGACY:
            z3i = get_z3_solver(z3i, af, args.assume_independence)
        else:  # solver in pysmt_solvers or solver is None
            si = get_smt_solver(si, af, solver)

        threshold = args.credulous_acceptance

        ca_timer = Timer(output_string="Checking credulous acceptance took")
        for argument in arguments:
            print("Checking credulous acceptance of argument", argument, "with threshold", threshold,
                  "under the following semantics:", repr_list(semantics))
            if solver == Z3_LEGACY:
                (acceptable, distribution) = tasks.check_credulous_acceptance_z3(z3i, semantics, argument, threshold)
            else:
                (acceptable, distribution) = tasks.check_credulous_acceptance(si, semantics, argument, threshold)

            if acceptable:
                print(argument, "is credulously accepted.")
                print("Satisfying distribution:")
                distribution.print_formatted(args.distribution_output_format)
            else:
                print(argument, "is not credulously accepted.")
        ca_timer.stop()

    if args.skeptical_acceptance is not None:
        if not af:
            arg_parser.error("an AF must be provided to check acceptance")
        if not arguments:
            arg_parser.error("no arguments given to check acceptance, provide one or more with -a")

        if solver in linear_solvers:
            arg_parser.error("skeptical acceptance checking not available with linear solver '" + solver + "'")
        elif solver == Z3_LEGACY:
            z3i = get_z3_solver(z3i, af, args.assume_independence)
        else:  # solver in pysmt_solvers or solver is None
            si = get_smt_solver(si, af, solver)

        threshold = args.skeptical_acceptance

        sa_timer = Timer(output_string="Checking skeptical acceptance took")
        for argument in arguments:
            print("Checking skeptical acceptance of argument", argument, "with threshold", threshold,
                  "under the following semantics:", repr_list(semantics))
            if solver == Z3_LEGACY:
                (acceptable, counterexample) = tasks.check_skeptical_acceptance_z3(z3i, semantics, argument, threshold)
            else:
                (acceptable, counterexample) = tasks.check_skeptical_acceptance(si, semantics, argument, threshold)
            if acceptable:
                print(argument, "is skeptically accepted.")
            else:
                print(argument, "is not skeptically accepted.")
                print("Counterexample:")
                counterexample.print_formatted(args.distribution_output_format)
        sa_timer.stop()

    if args.credulous_acceptance_all is not None:
        if not af:
            arg_parser.error("an AF must be provided to check acceptance")

        if solver in linear_solvers:
            arg_parser.error("credulous acceptance checking not available with linear solver '" + solver + "'")
        elif solver == Z3_LEGACY:
            z3i = get_z3_solver(z3i, af, args.assume_independence)
        else:  # solver in pysmt_solvers or solver is None
            si = get_smt_solver(si, af, solver)

        threshold = args.credulous_acceptance_all

        ca_timer = Timer(output_string="Checking credulous acceptance for all arguments took")
        print("Checking credulous acceptance of all arguments with threshold", threshold,
              "under the following semantics:", repr_list(semantics))
        accepted_args = []
        not_accepted_args = []
        for argument in af.get_nodes():
            if solver == Z3_LEGACY:
                acceptable, distribution = tasks.check_credulous_acceptance_z3(z3i, semantics, threshold, argument)
            else:
                acceptable, distribution = tasks.check_credulous_acceptance(si, semantics, threshold, argument)
            if acceptable:
                accepted_args.append(argument)
                # print(argument.name, "is credulously accepted.")
                # print("Satisfying distribution:")
                # distribution.print_formatted("SM")
            else:
                not_accepted_args.append(argument)
                # print(argument.name, "is not credulously accepted.")
        print("Credulously accepted:", accepted_args)
        print("Not credulously accepted:", not_accepted_args)
        ca_timer.stop()

    if args.skeptical_acceptance_all is not None:
        if not af:
            arg_parser.error("an AF must be provided to check acceptance")

        if solver in linear_solvers:
            arg_parser.error("sceptical acceptance checking not available with linear solver '" + solver + "'")
        elif solver == Z3_LEGACY:
            z3i = get_z3_solver(z3i, af, args.assume_independence)
        else:  # solver in pysmt_solvers or solver is None
            si = get_smt_solver(si, af, solver)

        threshold = args.skeptical_acceptance_all

        ca_timer = Timer(output_string="Checking skeptical acceptance for all arguments took")
        print("Checking skeptical acceptance of all arguments with threshold", threshold,
              "under the following semantics:", repr_list(semantics))
        accepted_args = []
        not_accepted_args = []
        for argument in af.get_nodes():
            if solver == Z3_LEGACY:
                acceptable, counterexample = tasks.check_skeptical_acceptance_z3(z3i, semantics, threshold, argument)
            else:
                acceptable, counterexample = tasks.check_skeptical_acceptance(si, semantics, threshold, argument)
            if acceptable:
                accepted_args.append(argument)
                # print(argument.name, "is skeptically accepted.")
            else:
                not_accepted_args.append(argument)
                # print(argument.name, "is not skeptically accepted.")
                # print("Counterexample:", counterexample)
                # counterexample.print_formatted("SM")
        print("Skeptically accepted:", accepted_args)
        print("Not skeptically accepted:", not_accepted_args)
        ca_timer.stop()

    main_timer.stop()

    # CSV friendly print of timing results
    # semantics_names = ",".join([semantics_obj.__class__.__name__ for semantics_obj in semantics])
    # timings = ";".join([str(timer.last_timing) for timer in Timer.all_timers])
    # # print(semantics_names + ";" + timings)
    # print(af.name + ";" + timings)


def assert_linear_semantics_availability(semantics: List[sem.Semantics], arg_parser):
    """
    Test availability of linear constraints for given semantics without constructing potentially expensive constraints.
    """
    dummy_af = DPGM()
    dummy_ls = LinearSolver(dummy_af)
    for s in semantics:
        dummy_s = s.__class__(dummy_af, complemented=s.complemented)
        try:
            dummy_s.get_linear_constraints(dummy_ls)
        except NotImplementedError:
            arg_parser.error("No linear constraints available (or implemented) for semantics '" + str(dummy_s) + "'")


def get_linear_solver(ls, af, solver_name) -> LinearSolver:
    if ls is None:
        ls_setup_timer = Timer(output_string="linear solver setup took")
        ls = LinearSolver(af, solver_name)
        ls_setup_timer.stop()
    return ls


def get_z3_solver(z3i, af, assume_independence=False) -> Z3Instance:
    if z3i is None:
        z3i_setup_timer = Timer(output_string="z3i setup took")
        z3i = Z3Instance(af, assume_independence)
        z3i_setup_timer.stop()
    return z3i


def get_smt_solver(si, af, solver_name) -> SolverInstance:
    if si is None:
        si_setup_timer = Timer(output_string="solver instance setup took")
        si = SolverInstance(af, solver_name=solver_name)
        si_setup_timer.stop()
    return si


def repr_list(s: List[str]):
    if len(s) > 0:
        return ", ".join(map(str, s))
    else:
        return "(none)"


def repr_list_i(s: List[str]):
    assert len(s) > 0
    if len(s) == 1:
        return str(s[0])
    else:
        return ", ".join(map(str, s[:-1])) + " and " + str(s[-1])


def plural(s: List[str]):
    if len(s) > 1:
        return "s"
    else:
        return ""


if __name__ == "__main__":
    main()



from typing import List, Tuple

import numpy as np
import pypoman
import pypoman.polyhedron
import cvxopt

from DPGM import Node
import DPGM as AF
from assignments import Assignment
import assignments as asg
from distribution import Distribution
from timer import Timer


# TODO:
#  - also handle equalities as constraints
#  - sparse matrix representation of constraints?


Constraint = Tuple[np.array, float]

DEFAULT_SOLVER = "glpk"


class LinearConstraintError(Exception):
    pass


def config_solver(solver: str):
    if solver is None:
        return config_solver(DEFAULT_SOLVER)
    if solver == "cvxopt":
        return solver
    if solver == "mosek":
        import mosek
        # suppress output of mosek solver
        cvxopt.solvers.options['mosek'] = {mosek.iparam.log: 0}
        return solver
    if solver == "glpk":
        # suppress output of glpk solver
        cvxopt.solvers.options['glpk'] = {'msg_lev': 'GLP_MSG_OFF'}
        return solver
    raise ValueError("Unknown solver: " + solver)


class LinearSolver:

    def __init__(self, af: AF, solver=DEFAULT_SOLVER):
        self.af = af
        self.solver = config_solver(solver)
        self.all_nodes: List[Node] = sorted(list(af.get_nodes()))
        self.dimension = 2**len(self.all_nodes)

        self.prob_dist_constraints: List[Constraint] = self.generate_prob_dist_constraints()

    def generate_prob_dist_constraints(self):
        constraints = []
        # each entry in the solution vector needs to be larger than zero
        for i in range(self.dimension):
            row = [0] * self.dimension
            row[i] = -1
            constraints.append((row, 0))
        # overall sum equals one
        constraints.append(([1] * self.dimension, 1))
        constraints.append(([-1] * self.dimension, -1))
        return constraints

    def constraint_matrix(self, constraints: List[Constraint]):
        matrix_A_list = []
        vector_b_list = []
        for (row, value) in self.prob_dist_constraints + constraints:
            matrix_A_list.append(row)
            vector_b_list.append(value)

        A = np.array(matrix_A_list)
        b = np.array(vector_b_list)

        # print("\nSystem of linear inequalities:")
        # print("A: \n" + str(A))
        # print("b:", b)
        # print("Solving: A x <= b")

        return A, b

    def constraint_matrix2(self, constraints: List[Constraint]):
        # build matrix directly without using self.prob_dist_constraints
        A = - np.identity(self.dimension)
        b = np.zeros(self.dimension)

        a1 = np.ones(self.dimension)
        a2 = - a1

        print("num constraints:", len(constraints))
        a_list, b_list = list(zip(*constraints))

        A = np.concatenate((A, [a1, a2], a_list))
        b = np.concatenate((b, [1, -1], b_list))

        return A, b

    def constraint_matrix_sparse(self, constraints: List[Constraint]):
        # build matrix directly without using self.prob_dist_constraints

        # sparse representation of diagonal matrix
        # [-1 0] x <= 0
        # [0 -1] x <= 0
        column_indices = list(range(self.dimension))
        row_indices = list(range(self.dimension))
        values = [-1] * self.dimension
        b = [0] * self.dimension
        current_row = self.dimension

        # [1 1] x <= 1
        values.extend([1] * self.dimension)
        column_indices.extend(range(self.dimension))
        row_indices.extend([current_row] * self.dimension)
        b.append(1)
        current_row += 1

        # [-1 -1 x <= -1
        values.extend([-1] * self.dimension)
        column_indices.extend(range(self.dimension))
        row_indices.extend([current_row] * self.dimension)
        b.append(-1)
        current_row += 1

        for (row, value) in constraints:
            if all([x == 0 for x in row]) and value == 0:
                # skip all-zero constraints like [0 0] x <= 0
                continue
            for i in range(self.dimension):
                if row[i] != 0:
                    values.append(int(row[i]))  # conversion to int necessary because the values originate from numpy
                    column_indices.append(i)
                    row_indices.append(current_row)
            b.append(value)
            current_row += 1

        A = cvxopt.spmatrix(values, row_indices, column_indices, tc='d')
        b = cvxopt.matrix(b, tc='d')
        return A, b

    def any_solution(self, constraints: List[Constraint]):
        c = [0] * self.dimension
        return self.optimal_distribution(c, constraints)

    def chebyshev_center(self, constraints: List[Constraint]):
        A, b = self.constraint_matrix(constraints)
        b = cvxopt.matrix(b, tc="d")
        try:
            x = pypoman.polyhedron.compute_chebyshev_center(A, b)
        except Exception as ex:
            if str(ex).startswith("Polytope is empty"):  # thrown by pypoman library
                return None
            else:
                raise ex
        return self.vector_to_distribution(x)

    def corner_distributions(self, constraints: List[Constraint]) -> List[Distribution]:
        A, b = self.constraint_matrix(constraints)

        # computes corners of solution space with A * x <= b
        vertices = pypoman.compute_polytope_vertices(A, b)

        # print("Simplified system:")
        # A, b = pypoman.compute_polytope_halfspaces(vertices)
        # print("A: \n" + str(A))
        # print("b:", b)

        # print("Vertices:", vertices)
        distributions = []
        for vector in vertices:
            distributions.append(self.vector_to_distribution(vector))
        return distributions

    def optimal_distribution(self, cost_vector, constraints: List[Constraint]):
        assert len(cost_vector) == self.dimension

        # A, b = self.constraint_matrix(constraints)
        # A = cvxopt.matrix(A, tc="d")
        # b = cvxopt.matrix(b, tc="d")
        A, b = self.constraint_matrix_sparse(constraints)
        c = cvxopt.matrix(cost_vector, tc="d")

        t = Timer(output_string=self.solver + " solver took")
        # x = pypoman.solve_lp(c, A, b, solver=solver)
        result = cvxopt.solvers.lp(c, A, b, solver=self.solver)
        t.stop()
        status = result['status']
        if status == "optimal":
            x = result['x']
            # print("sum:", sum(x))
            return self.vector_to_distribution(x)
        else:
            # print("Status:", status)
            return None

    def optimize_marginal_prob(self, nodes: List[Node], constraints: List[Constraint], mode="min") -> Distribution:
        """
        Compute the optimal distribution which minimizes (or maximizes) the marginal probability of one or more given
        arguments while observing the given constraints.

        :param nodes: a list of arguments
        :param constraints: a list of linear constraints
        :param mode: "min" (default) or "max"
        :return: a optimal distribution if the constraints are satisfiable
        """
        if mode == "min":
            sign = 1
        elif mode == "max":
            sign = -1
        else:
            raise ValueError("Unknown mode '" + mode + "'. Expected 'min' or 'max'.")
        assignments = []
        for node in nodes:
            assignments.extend(asg.get_node_assignments(node, self.all_nodes))
        cost_vector = sign * self.assignments_to_row(assignments)
        return self.optimal_distribution(cost_vector, constraints)

    def adm_degree(self):
        # todo: this does not completely follow the write-up as all nodes are considered, and not only those whose
        # marginal prob is positive. then again, this is probably not realizable as linear constraints
        rows = []
        for node in self.all_nodes:
            intersection_list = []
            for attacker in node.get_parents():
                union_set = set()
                for defender in attacker.get_parents():
                    if node != defender:
                        # noinspection PyTypeChecker
                        assignment: Assignment = ((node, True), (defender, True))
                    else:
                        assignment = ((node, True),)
                    defender_assignments = asg.sum_out(assignment, self.all_nodes)
                    union_set.update(defender_assignments)
                intersection_list.append(union_set)
            if intersection_list:
                assignments = list(set.intersection(*intersection_list))
                rows.append(self.assignments_to_row(assignments))
        return (1/len(self.all_nodes)) * sum(rows)

    def assignments_to_row(self, assignments: List[Assignment]):
        """
        [((A,False),(B,False)), ((A,True),(B,False))] -> (1 0 1 0)
        """
        row = np.array([0] * self.dimension)
        for assignment in assignments:
            row[asg.to_id(assignment)] = 1.0
        return row

    def constraints_assignments_less_or_equal_value(
            self, assignments: List[Assignment], value) -> List[Constraint]:
        """
        Generate constraints such that the sum of the given assignments is less or equal to the given value.
        """
        row = self.assignments_to_row(assignments)
        return [(row, value)]

    def constraints_assignments_greater_or_equal_value(
            self, assignments: List[Assignment], value) -> List[Constraint]:
        """
        Generate constraints such that the sum of the given assignments is greater or equal to the given value.
        """
        row = -1 * self.assignments_to_row(assignments)
        return [(row, -value)]

    def constraints_assignments_equal_value(
            self, assignments: List[Assignment], value) -> List[Constraint]:
        """
        Generate constraints such that the sum of the given assignments equals the given value.
        """
        return self.constraints_assignments_less_or_equal_value(assignments, value) + \
            self.constraints_assignments_greater_or_equal_value(assignments, value)

    def constraints_assignments_less_or_equal_assignments(
            self, assignments_a: List[Assignment], assignments_b: List[Assignment]) -> List[Constraint]:
        """
        Generate constraints such that the sum of the first assignments is less or equal to the sum of the second
        assignments.
        """
        row_a = self.assignments_to_row(assignments_a)
        row_b = self.assignments_to_row(assignments_b)
        return [(row_a - row_b, 0)]

    def constraints_assignments_greater_or_equal_assignments(
            self, assignments_a: List[Assignment], assignments_b: List[Assignment]) -> List[Constraint]:
        """
        Generate constraints such that the sum of the first assignments is greater or equal to the sum of the second
        assignments.
        """
        row_a = self.assignments_to_row(assignments_a)
        row_b = self.assignments_to_row(assignments_b)
        return [(row_b - row_a, 0)]

    def constraints_assignments_equal_assignments(
            self, assignments_a: List[Assignment], assignments_b: List[Assignment]) -> List[Constraint]:
        """
        Generate constraints such that the sum of the first assignments equals the sum of the second assignments.
        """
        return self.constraints_assignments_less_or_equal_assignments(assignments_a, assignments_b) + \
            self.constraints_assignments_greater_or_equal_assignments(assignments_a, assignments_b)

    def constraints_node_less_or_equal_value(self, node: Node, value) -> List[Constraint]:
        node_full_assignments = asg.get_node_assignments(node, self.all_nodes)
        return self.constraints_assignments_less_or_equal_value(node_full_assignments, value)

    def constraints_node_greater_or_equal_value(self, node: Node, value) -> List[Constraint]:
        node_full_assignments = asg.get_node_assignments(node, self.all_nodes)
        return self.constraints_assignments_greater_or_equal_value(node_full_assignments, value)

    def constraints_node_equals_value(self, node: Node, value) -> List[Constraint]:
        node_full_assignments = asg.get_node_assignments(node, self.all_nodes)
        return self.constraints_assignments_equal_value(node_full_assignments, value)

    def get_node_value(self, node: Node, distribution_vector):
        """
        Return the marginal probability of the given node under the distribution.
        """
        prob_sum = 0
        for a in asg.get_node_assignments(node, self.all_nodes):
            prob_sum += distribution_vector[asg.to_id(a)]
        return prob_sum

    def vector_to_distribution(self, distribution_vector):
        distribution = dict()
        for i in range(self.dimension):
            value = distribution_vector[i]
            distribution[asg.from_id(i, self.all_nodes)] = value
        return Distribution(distribution, self.all_nodes)



from typing import Set, Tuple, Dict, List


DEFAULT_EDGE_TYPE = "attack"


class Node:
    """A node in a directed probabilistic graphical model"""
    __default_node_prefix = ""  # for nodes without a given name, the name is this prefix followed by the node's ID

    def __init__(self, node_id, name: str = ""):
        self.id = node_id
        self.dpgm = None
        if not name:
            self.name = self.__default_node_prefix + str(self.id)
        else:
            self.name = name

        self.all_parents: Set[Node] = set()
        self.parents: Dict[str, Set[Node]] = dict()  # parents by edge type
        self.parents_edges = dict()  # Dict[Node.id, Edge]

        self.all_children: Set[Node] = set()
        self.children:  Dict[str, Set[Node]] = dict()  # children by edge type
        self.children_edges = dict()  # Dict[Node.id, Edge]

        self.value = None  # float
        self.interval = None  # Tuple[float, float]

    def add_value(self, value: float):
        self.value = value

    def add_interval(self, interval: Tuple[float, float]):
        self.interval = interval

    def is_initial(self):
        return len(self.all_parents) == 0

    def get_parents(self, edge_type=DEFAULT_EDGE_TYPE):
        if edge_type in self.parents.keys():
            return self.parents[edge_type]
        else:
            return set()

    def get_children(self, edge_type=DEFAULT_EDGE_TYPE):
        if edge_type in self.children.keys():
            return self.children[edge_type]
        else:
            return set()

    def get_parent_edge(self, node_from):
        if node_from.id in self.parents_edges:
            return self.parents_edges[node_from.id]
        else:
            raise ValueError("Node '" + self.name + "' has no parent edge from node '" + node_from.name + "'.")

    def get_child_edge(self, node_to):
        if node_to.id in self.children_edges:
            return self.children_edges[node_to.id]
        else:
            raise ValueError("Node '" + self.name + "' has no child edge to node '" + node_to.name + "'.")

    def __eq__(self, other):
        return self.id == other.id

    def __lt__(self, other):
        return self.name < other.name

    def __gt__(self, other):
        return self.name > other.name

    def __hash__(self):
        return hash(self.id)

    def __repr__(self):
        return self.name
        # return "Node(" + self.id + ": " + self.name + ")"

    def __str__(self):
        return self.name


class Edge:
    """A directed edge in a directed probabilistic graphical model"""
    __default_label_prefix = "s"  # default label "s_A_B" for an edge from node A to node B

    def __init__(self, node_from, node_to, label="", edge_type=DEFAULT_EDGE_TYPE):
        self.node_from: Node = node_from
        self.node_to: Node = node_to
        self.label: str = label
        self.value = None
        self.interval = None
        self.type = edge_type  # e.g. "attack", "support"
        if self.label == "":
            self.label = self.__default_label_prefix + '_' + self.node_from.name + '_' + self.node_to.name

        self.node_from.all_children.add(self.node_to)
        if self.type not in self.node_from.children.keys():
            self.node_from.children[self.type] = set()
        self.node_from.children[self.type].add(self.node_to)
        self.node_from.children_edges[node_to.id] = self

        self.node_to.all_parents.add(self.node_from)
        if self.type not in self.node_to.parents.keys():
            self.node_to.parents[self.type] = set()
        self.node_to.parents[self.type].add(self.node_from)
        self.node_to.parents_edges[node_from.id] = self

    def add_value(self, value: float):
        self.value = value

    def add_interval(self, interval: Tuple[float, float]):
        self.interval = interval

    def __str__(self):
        return str(self.node_from) + "->" + str(self.node_to)


class DPGM:
    """A class representing a directed probabilistic graphical model"""
    def __init__(self, name: str = "", path: str = ""):
        self.name = name
        self.path = path
        self.nodes: Dict[str, Node] = dict()
        self.all_edges: Set[Edge] = set()
        self.edges: Dict[str, Set[Edge]] = dict()  # edges by type

    def add_node(self, node: Node):
        self.nodes[node.id] = node
        node.dpgm = self

    def add_edge(self, edge: Edge):
        self.all_edges.add(edge)
        if edge.type not in self.edges.keys():
            self.edges[edge.type] = set()
        self.edges[edge.type].add(edge)

    def get_node(self, node_id: str):
        return self.nodes[node_id]

    def get_nodes(self):
        return self.nodes.values()

    def get_node_by_name(self, name: str):
        for node in self.get_nodes():
            if node.name == name:
                return node
        raise ValueError("No node with name '" + name + "' found.")

    def get_initial_nodes(self):
        return set([node for node in self.nodes.values() if node.is_initial()])

    def get_edges(self):
        return self.all_edges

    def __str__(self):
        s = str(self.name)
        s += "\n" + str(len(self.nodes)) + " nodes: " + ', '.join([node.name for node in self.nodes.values()])
        s += "\n" + str(len(self.all_edges)) + " edges: " \
             + ', '.join([edge.node_from.name + '->' + edge.node_to.name for edge in self.all_edges])
        return s


def parents(nodes: List[Node], edge_type=DEFAULT_EDGE_TYPE) -> List[Node]:
    """
    Return the union of the nodes' parents w.r.t. to a an edge type
    :param nodes: A list of nodes
    :param edge_type: The type of the considered edges
    :return: The sorted list of the parents
    """
    parents_list = []
    for node in nodes:
        parents_list.extend([p for p in node.get_parents(edge_type) if p not in parents_list])
    parents_list.sort()
    return parents_list



from typing import List, Tuple, Any, Set

from DPGM import Node

Assignment = Tuple[Tuple[Node, bool]]  # type alias


# class Assignment:
#     def __init__(self, collection):
#
#         # self.assignment_dict = dict()
#         # for (node, b) in collection:
#         #     self.assignment_dict[node] = b
#
#         self.assignment = tuple(sorted(list(collection)))
#
#     def update(self, node: Node, b: bool):
#         self.assignment_dict[node] = b
#
#     def get_nodes(self) -> List[Node]:
#         return [node for (node, _) in self.assignment]
#
#     def __hash__(self):
#         return self.assignment.__hash__()
#
#     def __eq__(self, other):
#         return self.assignment == other.assignment
#
#     def __repr__(self):
#         return repr(self.assignment)


def generate(ar: list, br: list = (False, True)) -> List[Assignment]:
    """
    Generates all possible assignments of the elements in list 'ar' to elements of list 'br'.
    E.g. ar = [x,y], br = [0,1] -> [[(x,0),(y,0)], [(x,0),(y,1)], [(x,1),(y,0)], [(x,1),(y,1)]]
    """
    combinations: List[List[Tuple[Node, bool]]] = [[]]

    for a in ar:
        next_combinations = []
        for c in combinations:
            for b in br:
                new_c = c.copy()
                new_c.append((a, b))
                next_combinations.append(new_c)
        combinations = next_combinations
    assignments = []
    for c in combinations:
        assignments.append(tuple(c))
    return assignments


def repr_assignment(a: Assignment) -> str:
    strings = []
    for (node, b) in a:
        strings.append((" " if b else "-") + str(node))
    return ",".join(strings)


def to_id(assignment: Assignment):
    """
    Given an assignment like [("A", False), ("B", True), ("C", False)], converts the assignment to a binary string "010"
    and returns the decimal value of it.
    """
    return int("".join(["1" if b else "0" for (_, b) in assignment]), 2)


def from_id(assignment_id: int, nodes: List[Node]) -> Assignment:
    """
    Reverse function of to_id(), i.e., given ID 2 and nodes "ABC", we yield
    [("A", False), ("B", True), ("C", False)].
    """
    assert assignment_id < 2**len(nodes)
    assignment: List[Tuple[Node, bool]] = \
        [(node, c == "1") for (node, c) in zip(nodes, format(assignment_id, "0" + str(len(nodes)) + "b"))]
    return tuple(assignment)


def to_extension(assignment: Assignment):
    """
    [(A,True),(B,False),(C,True)] -> {A,C}
    """
    return set([node for (node, b) in assignment if b])


def from_extension(extension, nodes: List[Node]) -> Assignment:
    """
    Extension {A,C} and nodes [A,B,C] yield [(A,True),(B,False),(C,True)].
    """
    return tuple([(node, node in extension) for node in nodes])


def sum_out(part_assignment: Assignment, all_nodes: List[Node]) -> List[Assignment]:
    """
    Given a partial assignment over some subset of all_nodes, return all full assignments agreeing with the given one.
    E.g., [(A,True),(B,False)], [A,B,C] -> [[(A,True),(B,False),(C,False)], [(A,True),(B,False)(C,True)].

    :param part_assignment: the partial assignment
    :param all_nodes: the list of all nodes
    :returns: the list of all full assignments agreeing with part_assignment
    """
    part_assignment = list(part_assignment)
    part_assignment_args = [arg for (arg, _) in part_assignment]
    remaining_args = [arg for arg in all_nodes if arg not in part_assignment_args]
    assignments_list = []
    for assignment in generate(remaining_args):
        full_assignment = part_assignment + list(assignment)
        full_assignment.sort()
        assignments_list.append(tuple(full_assignment))
    return assignments_list


def get_node_assignments(node: Node, nodes: List[Node], node_assignment=True) -> List[Assignment]:
    """
    Return the list of all assignments over `nodes` under which the given node has the given node assignment.
    """
    return sum_out(((node, node_assignment),), nodes)


def positive_cnf_assignments(clauses: List[List[Any]], all_nodes: List[Any]) -> List[Assignment]:
    """
    Given a positive CNF (i.e. without any negation in front of literals), return all full assignments which satisfy the
    CNF. E.g., for  (A OR B) AND (B OR C), given as [[A, B], [B, C]], return (A B C), (A B -C), (-A B C), -(A B -C) and
    (A -B C).

    :param clauses: A list of the conjunctive clauses of the positive CNF
    :param all_nodes: all nodes to be considered, i.e. all nodes of the AF in most cases
    :return: A list of all full assignments which satisfy the CNF
    """
    candidates: List[Set[Tuple[Node, bool]]] = [set()]
    given_arguments = set()
    for clause in clauses:
        given_arguments.update(clause)
        clause_assignments = generate(clause, [False, True])
        # Clauses contain disjunctive literals, so a true assignment is any assignment which assigns true at least
        # once. Filter out the first result from 'assignments' function because this assigns False to all literals.
        true_clause_assignments = clause_assignments[1:]

        new_candidates = []
        for candidate in candidates:
            # e.g. candidate = [(A, True), (B, False)]
            for true_clause_assignment in true_clause_assignments:
                # e.g. true_clause_assignment = [(B, False), (C, True)]
                # check whether the arguments appearing in both assignments are assigned the same value
                skip = False
                for (node, b) in true_clause_assignment:
                    if (node, not b) in candidate:
                        skip = True
                        break
                # if so, add the new arguments to the candidate
                if not skip:
                    new_candidate = candidate.copy()
                    new_candidate.update(true_clause_assignment)
                    new_candidates.append(new_candidate)
        candidates = new_candidates

    # arguments not present in the cnf can take on any value, so each previous candidate is extended by all possible
    # assignments of the remaining arguments
    remaining_arguments = set(all_nodes).difference(given_arguments)
    remaining_assignments = generate(list(remaining_arguments))
    final_candidates: List[Assignment] = []
    for candidate in candidates:
        for remaining_assignment in remaining_assignments:
            final_candidate = candidate.copy()
            final_candidate.update(remaining_assignment)
            final_candidates.append(tuple(sorted(list(final_candidate))))
    return final_candidates



from fractions import Fraction
from typing import List

import pysmt.typing as types
from pysmt.shortcuts import Equals, Symbol, Real, Plus, GE, LE, And
import pysmt.shortcuts as smt
from pysmt.solvers.solver import Model
from pysmt.smtlib.parser import SmtLibParser


from assignments import Assignment
import assignments
import DPGM as AF
from distribution import Distribution


class SolverInstance:
    def __init__(self, af: AF, solver_name=None):
        self.af = af
        self.dimension = 2**len(af.get_nodes())
        self.solver_name = solver_name

        self.full_assignments: List[Assignment] = []
        self.full_assignment_symbols = dict()
        self.assignment_symbols = dict()
        self.constraints = []  # internal constraints that are always enforced by being added to the solver by default
        self.solver = None

        self._generate_full_assignment_symbols()

    def _generate_full_assignment_symbols(self):
        if not self.full_assignments:
            self.full_assignments = assignments.generate(sorted(self.af.get_nodes()))
        for assignment in self.full_assignments:
            name = symbol_name(assignment)
            symbol = Symbol(name, types.REAL)
            self.full_assignment_symbols[assignment] = symbol
            self.constraints.append(GE(symbol, Real(0)))
            # self.constraints.append(LE(symbol, Real(1)))
        # sum of all full assignments equals one
        self.constraints.append(Equals(Plus(self.full_assignment_symbols.values()), Real(1)))

    def get_symbol(self, assignment: Assignment):
        assert assignment, "empty assignment"

        if assignment in self.full_assignment_symbols:
            return self.full_assignment_symbols[assignment]
        elif assignment in self.assignment_symbols:
            return self.assignment_symbols[assignment]
        else:
            return self.generate_symbol(assignment)

    def get_node_symbol(self, node: AF.Node):
        return self.get_symbol(tuple([(node, True)]))

    def generate_symbol(self, assignment: Assignment):
        """Generate a symbol and constraints for the probability of a partial assignment."""
        assert len(assignment) < self.dimension
        name = symbol_name(assignment)
        symbol = Symbol(name, types.REAL)
        self.assignment_symbols[assignment] = symbol
        self._add_internal_constraint(GE(symbol, Real(0)))
        self._add_internal_constraint(LE(symbol, Real(1)))
        # express partial assignment as sum of full assignments
        full_assignments = assignments.sum_out(assignment, sorted(self.af.get_nodes()))
        constraint = Equals(symbol, Plus([self.full_assignment_symbols[asg] for asg in full_assignments]))
        self._add_internal_constraint(constraint)
        return symbol

    def distribution_from_model(self, model: Model) -> Distribution:
        """
        Get a distribution object from a model.
        """
        distribution_dict = dict()
        for assignment in self.full_assignments:
            distribution_dict[assignment] = self.value_from_model(assignment, model)
            # distribution_dict[assignment] = self.value_from_model_fraction(assignment, model)
        distribution = Distribution(distribution_dict, self.af.get_nodes())
        distribution.model = model
        return distribution

    def value_from_model(self, assignment: Assignment, model: Model):
        symbol = self.get_symbol(assignment)
        constant = model.get_value(symbol)
        if constant.is_real_constant() or constant.is_int_constant():
            return float(constant.constant_value())
        else:
            raise ValueError("Constant is neither INT or REAL: " + str(constant))

    def value_from_model_fraction(self, assignment: Assignment, model: Model) -> Fraction:
        symbol = self.get_symbol(assignment)
        constant = model.get_value(symbol)
        if constant.is_real_constant() or constant.is_int_constant():
            value = constant.constant_value()
            if isinstance(value, Fraction):
                return value
            else:
                return constant.algebraic_approx_value()
        else:
            raise ValueError("Constant is neither INT or REAL: " + str(constant))

    def is_sat(self, constraints):
        solver = self.get_solver()
        return solver.is_sat(And(constraints))

    def get_model(self, constraints):
        if self.is_sat(constraints):
            return self.solver.get_model()

    def get_solver(self) -> smt.Solver():
        if not self.solver:
            # create a new solver and add internal constraints
            if self.solver_name:
                self.solver = smt.Solver(name=self.solver_name)
            else:
                self.solver = smt.Solver()
            # print("basic constraints: ", self.constraints)
            self.solver.add_assertion(And(self.constraints))
        return self.solver

    def _add_internal_constraint(self, constraint):
        self.constraints.append(constraint)
        if self.solver:
            self.solver.add_assertion(constraint)

    def add_constraints_from_smt(self, smt_file):
        parser = SmtLibParser()
        try:
            script = parser.get_script(smt_file)
        except AttributeError:
            raise ValueError("Invalid SMT file. (Are all symbols declared?)")
        if not script.contains_command("check-sat"):
            script.add("check-sat", "")
        # print(script)
        for smt_symbol in script.get_declared_symbols():
            asg = symbol_name_to_assignment(str(smt_symbol), self.af)
            symbol = self.get_symbol(asg)
            if smt_symbol != symbol:
                raise ValueError("Invalid symbol name in SMT file: '" + str(smt_symbol)
                                 + "'. Did you mean + '" + str(symbol) + "'?")
        constraint = script.get_strict_formula()
        print("SMT constraint: ", constraint)
        self.constraints.append(constraint)


def symbol_name(assignment: Assignment) -> str:
    """
    Generates a fixed name for any assignment.
    Expects a list of tuples of nodes matched to True or False, i.e. [(node_B, False), (node_C, True)].
    Generates a name for the probability of the conjunction of the nodes, i.e. 'p_nB_C'.
    The order is determined by node IDs.
    """
    # assert list(assignment) == sorted(list(assignment))  # TODO: enforce sortedness as invariant?
    assignment = sorted(list(assignment))
    name = "p"
    for (arg, b) in assignment:
        name += "_"
        if not b:
            name += "n"
        name += arg.name
    return name


def symbol_name_to_assignment(name: str, af: AF) -> Assignment:
    """
    Inverse of `symbol_name` function.
    """
    split = name.split("_")
    assert split[0] == "p"

    assignment = []
    for i in range(1, len(split)):
        token = split[i]
        b = True
        if token.startswith("n"):
            b = False
            token = token[1:]
        node: AF.Node = af.get_node_by_name(token)
        assignment.append((node, b))
    return tuple(assignment)



import textwrap
from fractions import Fraction
from typing import List, Type, Set, Tuple

import z3
from pysmt.shortcuts import Not, And, Or, Implies, Iff, Equals, GE, GT, LE, Minus, Plus, Real, Bool

import assignments as asg
import classical_semantics as classical
from DPGM import DPGM, Node
from solver_instance import SolverInstance
from z3_instance import Z3Instance
from linear_solver import LinearSolver, Constraint


MAX_FRACTION_DENOMINATOR = 10000
CONSTRAINT_LOGGING = False
LOGGING_MAX_WIDTH = 120


class Semantics:
    def __init__(self, af: DPGM, complemented=False):
        self.af = af
        self.all_nodes = sorted(list(af.get_nodes()))
        self.constraints = None  # smt-lib constraints
        self.constraints_z3 = None
        self.constraints_linear = None
        self.complemented = complemented  # if this flag is set, the semantics' constraints are negated

    def generate_constraints(self, si: SolverInstance):
        # implemented by subclasses
        raise NotImplementedError

    def generate_z3_constraints(self, z3i: Z3Instance):
        # implemented by subclasses
        raise NotImplementedError

    def generate_linear_constraints(self, ls: LinearSolver):
        # implemented by subclasses
        raise NotImplementedError

    def get_constraints(self, si: SolverInstance) -> List:
        if self.constraints is None:
            self.constraints = []
            self.generate_constraints(si)
        if not self.complemented:
            return self.constraints
        else:
            return [Not(And(self.constraints))]

    def get_z3_constraints(self, z3i: Z3Instance) -> List:
        if self.constraints_z3 is None:
            self.constraints_z3 = []
            self.generate_z3_constraints(z3i)
        if not self.complemented:
            return self.constraints_z3
        else:
            return [z3.Not(z3.And(self.constraints_z3), ctx=z3i.context)]

    def get_linear_constraints(self, ls: LinearSolver) -> List[Constraint]:
        if self.complemented:
            raise NotImplementedError("Complement semantics not supported for linear constraints.")
        if self.constraints_linear is None:
            self.constraints_linear = []
            self.generate_linear_constraints(ls)
        return self.constraints_linear

    def add_constraint(self, constraint, node=None, edge=None):
        self.constraints.append(constraint)
        if CONSTRAINT_LOGGING:
            if node:
                print(str(self) + ": " + str(node) + ": " + textwrap.shorten(str(constraint), LOGGING_MAX_WIDTH))
            elif edge:
                print(str(self) + ": " + str(edge) + ": " + textwrap.shorten(str(constraint), LOGGING_MAX_WIDTH))
            else:
                print(str(self) + ": " + textwrap.shorten(str(constraint), LOGGING_MAX_WIDTH))

    def add_z3_constraint(self, constraint, node=None, edge=None):
        self.constraints_z3.append(constraint)
        if CONSTRAINT_LOGGING:
            if node:
                print(str(self) + ": " + str(node) + ": " + textwrap.shorten(str(constraint), LOGGING_MAX_WIDTH))
            elif edge:
                print(str(self) + ": " + str(edge) + ": " + textwrap.shorten(str(constraint), LOGGING_MAX_WIDTH))
            else:
                print(str(self) + ": " + textwrap.shorten(str(constraint), LOGGING_MAX_WIDTH))

    def __str__(self):
        name = self.__class__.__name__
        if self.complemented:
            name = "co-" + name
        return name


####################
# trivial semantics
####################

class Min(Semantics):
    """
    Minimality semantics: All nodes must hold with probability 0.
    """
    def generate_constraints(self, si: SolverInstance):
        for node in self.all_nodes:
            node_symbol = si.get_node_symbol(node)
            self.add_constraint(Equals(node_symbol, Real(0)), node=node)

    def generate_z3_constraints(self, z3i: Z3Instance):
        for node in self.all_nodes:
            node_prob = z3i.get_node_prob_var(node)
            self.add_z3_constraint(node_prob == 0, node=node)

    def generate_linear_constraints(self, ls: LinearSolver):
        for node in self.all_nodes:
            self.constraints_linear.extend(ls.constraints_node_equals_value(node, 0))


class Neu(Semantics):
    """
    Neutrality semantics: All nodes must hold with probability 0.5.
    """
    def generate_constraints(self, si: SolverInstance):
        for node in self.all_nodes:
            node_symbol = si.get_node_symbol(node)
            self.add_constraint(Equals(node_symbol, Real(0.5)), node=node)

    def generate_z3_constraints(self, z3i: Z3Instance):
        for node in self.all_nodes:
            node_prob = z3i.get_node_prob_var(node)
            self.add_z3_constraint(node_prob == 0.5, node=node)

    def generate_linear_constraints(self, ls: LinearSolver):
        for node in self.all_nodes:
            self.constraints_linear.extend(ls.constraints_node_equals_value(node, 0.5))


class Max(Semantics):
    """
    Maximality semantics: All nodes must hold with probability 1.
    """
    def generate_constraints(self, si: SolverInstance):
        for node in self.all_nodes:
            node_symbol = si.get_node_symbol(node)
            self.add_constraint(Equals(node_symbol, Real(1)), node=node)

    def generate_z3_constraints(self, z3i: Z3Instance):
        for node in self.all_nodes:
            node_prob = z3i.get_node_prob_var(node)
            self.add_z3_constraint(node_prob == 1, node=node)

    def generate_linear_constraints(self, ls: LinearSolver):
        for node in self.all_nodes:
            self.constraints_linear.extend(ls.constraints_node_equals_value(node, 1))


class Dirac(Semantics):
    """
    Dirac semantics: All nodes must hold with either probability 0 or 1.
    """
    def generate_constraints(self, si: SolverInstance):
        for node in self.all_nodes:
            node_symbol = si.get_node_symbol(node)
            self.add_constraint(Or(Equals(node_symbol, Real(0)), Equals(node_symbol, Real(1))), node=node)

    def generate_z3_constraints(self, z3i: Z3Instance):
        for node in self.all_nodes:
            node_prob = z3i.get_node_prob_var(node)
            self.add_z3_constraint(z3.Or(node_prob == 0, node_prob == 1), node=node)


class Ter(Semantics):
    """
    Ternary semantics: All nodes must hold with either probability 0, 0.5 or 1.
    """
    def generate_constraints(self, si: SolverInstance):
        for node in self.all_nodes:
            node_symbol = si.get_node_symbol(node)
            self.add_constraint(Or(Equals(node_symbol, Real(0)),
                                   Equals(node_symbol, Real(0.5)),
                                   Equals(node_symbol, Real(1))), node=node)

    def generate_z3_constraints(self, z3i: Z3Instance):
        for node in self.all_nodes:
            node_prob = z3i.get_node_prob_var(node)
            self.add_z3_constraint(z3.Or(node_prob == 0, node_prob == 0.5, node_prob == 1), node=node)


################################
# semantics by Hunter and Thimm
################################

class Fou(Semantics):
    """
    Foundedness semantics: Initial nodes must hold with probability 1.
    """
    def generate_constraints(self, si: SolverInstance):
        for node in self.af.get_initial_nodes():
            node_symbol = si.get_node_symbol(node)
            self.add_constraint(Equals(node_symbol, Real(1)), node=node)

    def generate_z3_constraints(self, z3i: Z3Instance):
        for node in self.af.get_initial_nodes():
            node_prob = z3i.get_node_prob_var(node)
            self.add_z3_constraint(node_prob == 1, node=node)

    def generate_linear_constraints(self, ls: LinearSolver):
        for node in self.af.get_initial_nodes():
            self.constraints_linear.extend(ls.constraints_node_equals_value(node, 1))


class SFou(Semantics):
    """
    Semi-Foundedness semantics: Initial nodes must hold with probability >= 0.5.
    """
    def generate_constraints(self, si: SolverInstance):
        for node in self.af.get_initial_nodes():
            node_symbol = si.get_node_symbol(node)
            self.add_constraint(GE(node_symbol, Real(0.5)), node=node)

    def generate_z3_constraints(self, z3i: Z3Instance):
        for node in self.af.get_initial_nodes():
            node_prob = z3i.get_node_prob_var(node)
            self.add_z3_constraint(node_prob >= 0.5, node=node)

    def generate_linear_constraints(self, ls: LinearSolver):
        for node in self.af.get_initial_nodes():
            self.constraints_linear.extend(ls.constraints_node_greater_or_equal_value(node, 0.5))


class Inv(Semantics):
    """
    Involution semantics: P(A) = 1 - P(B) for all attacks A -> B.
    """
    def generate_constraints(self, si: SolverInstance):
        for edge in self.af.get_edges():
            symbol_node_from = si.get_node_symbol(edge.node_from)
            symbol_node_to = si.get_node_symbol(edge.node_to)
            self.add_constraint(Equals(symbol_node_from, Minus(Real(1), symbol_node_to)), edge=edge)

    def generate_z3_constraints(self, z3i: Z3Instance):
        for edge in self.af.get_edges():
            prob_node_from = z3i.get_node_prob_var(edge.node_from)
            prob_node_to = z3i.get_node_prob_var(edge.node_to)
            self.add_z3_constraint(prob_node_from == 1 - prob_node_to, edge=edge)

    def generate_linear_constraints(self, ls: LinearSolver):
        for edge in self.af.get_edges():
            node_from_assignments = asg.get_node_assignments(edge.node_from, self.all_nodes)
            node_to_false_assignments = asg.get_node_assignments(edge.node_to, self.all_nodes,
                                                                 node_assignment=False)
            c = ls.constraints_assignments_equal_assignments(node_from_assignments, node_to_false_assignments)
            self.constraints_linear.extend(c)


class Rat(Semantics):
    """
    Rationality semantics: P(A) > 0.5 implies P(B) <= 0.5 for all attacks A -> B.
    """
    def generate_constraints(self, si: SolverInstance):
        for edge in self.af.get_edges():
            symbol_node_from = si.get_node_symbol(edge.node_from)
            symbol_node_to = si.get_node_symbol(edge.node_to)
            constraint = Implies(GT(symbol_node_from, Real(0.5)), LE(symbol_node_to, Real(0.5)))
            self.add_constraint(constraint, edge=edge)

    def generate_z3_constraints(self, z3i: Z3Instance):
        for edge in self.af.get_edges():
            prob_node_from = z3i.get_node_prob_var(edge.node_from)
            prob_node_to = z3i.get_node_prob_var(edge.node_to)
            constraint = z3.Implies(prob_node_from > 0.5, prob_node_to <= 0.5, ctx=z3i.context)
            self.add_z3_constraint(constraint, edge=edge)


class Coh(Semantics):
    """
    Coherency semantics: P(A) <= 1 - P(B) for all attacks A -> B.
    """
    def generate_constraints(self, si: SolverInstance):
        for edge in self.af.get_edges():
            symbol_node_from = si.get_node_symbol(edge.node_from)
            symbol_node_to = si.get_node_symbol(edge.node_to)
            constraint = LE(symbol_node_from, Minus(Real(1), symbol_node_to))
            self.add_constraint(constraint, edge=edge)

    def generate_z3_constraints(self, z3i: Z3Instance):
        for edge in self.af.get_edges():
            prob_node_from = z3i.get_node_prob_var(edge.node_from)
            prob_node_to = z3i.get_node_prob_var(edge.node_to)
            self.add_z3_constraint(prob_node_from <= 1 - prob_node_to, edge=edge)

    def generate_linear_constraints(self, ls: LinearSolver):
        for edge in self.af.get_edges():
            node_from_assignments = asg.get_node_assignments(edge.node_from, self.all_nodes)
            node_to_false_assignments = asg.get_node_assignments(edge.node_to, self.all_nodes,
                                                                 node_assignment=False)
            c = ls.constraints_assignments_less_or_equal_assignments(node_from_assignments, node_to_false_assignments)
            self.constraints_linear.extend(c)


class Opt(Semantics):
    """
    Optimism semantics: P(A) >= 1 - Sum P(B) for all attackers B of A. For initial arguments A, this entails P(A) = 1.
    """
    def generate_constraints(self, si: SolverInstance):
        for node in self.all_nodes:
            node_symbol = si.get_node_symbol(node)
            if node.is_initial():
                self.add_constraint(Equals(node_symbol, Real(1)), node=node)
                continue
            summation = Plus([si.get_node_symbol(attacker) for attacker in node.get_parents()])
            constraint = GE(node_symbol, Minus(Real(1), summation))
            self.add_constraint(constraint, node=node)

    def generate_z3_constraints(self, z3i: Z3Instance):
        for node in self.all_nodes:
            node_prob = z3i.get_node_prob_var(node)
            if node.is_initial():
                self.add_z3_constraint(node_prob == 1, node=node)
                continue
            summation = z3.Sum([z3i.get_node_prob_var(attacker) for attacker in node.get_parents()])
            self.add_z3_constraint(node_prob >= 1 - summation, node=node)

    def generate_linear_constraints(self, ls: LinearSolver):
        all_nodes = self.all_nodes
        for node in all_nodes:
            if node.is_initial():
                c = ls.constraints_node_equals_value(node, 1)
                self.constraints_linear.extend(c)
                continue
            # p_A >= 1 - Sum p_B is equivalent to p_A + Sum p_B >= 1 and -p_A + (-1) * Sum p_B <= -1
            sum_nodes = [node] + list(node.get_parents())
            rows = [ls.assignments_to_row(asg.get_node_assignments(n, all_nodes)) for n in sum_nodes]
            row = -1 * sum(rows)
            self.constraints_linear.append((row, -1))


class SOpt(Semantics):
    """
    Semi-optimism semantics: P(A) >= 1 - Sum P(B) for all non-initial arguments A with attackers B.
    """
    def generate_constraints(self, si: SolverInstance):
        for node in self.all_nodes:
            node_symbol = si.get_node_symbol(node)
            if node.is_initial():
                continue
            summation = Plus([si.get_node_symbol(attacker) for attacker in node.get_parents()])
            constraint = GE(node_symbol, Minus(Real(1), summation))
            self.add_constraint(constraint, node=node)

    def generate_z3_constraints(self, z3i: Z3Instance):
        for node in self.all_nodes:
            if node.is_initial():
                continue
            node_prob = z3i.get_node_prob_var(node)
            summation = z3.Sum([z3i.get_node_prob_var(attacker) for attacker in node.get_parents()])
            self.add_z3_constraint(node_prob >= 1 - summation, node=node)

    def generate_linear_constraints(self, ls: LinearSolver):
        all_nodes = self.all_nodes
        for node in all_nodes:
            if node.is_initial():
                continue
            sum_nodes = [node] + list(node.get_parents())
            rows = [ls.assignments_to_row(asg.get_node_assignments(n, all_nodes)) for n in sum_nodes]
            row = -1 * sum(rows)
            self.constraints_linear.append((row, -1))


class Jus(Semantics):
    """
    Justifiability semantics: Coherency and optimism constraints hold.
    """
    def generate_constraints(self, si: SolverInstance):
        opt = Opt(self.af)
        coh = Coh(self.af)
        self.constraints.extend(opt.get_constraints(si) + coh.get_constraints(si))

    def generate_z3_constraints(self, z3i: Z3Instance):
        opt = Opt(self.af)
        coh = Coh(self.af)
        self.constraints_z3.extend(opt.get_z3_constraints(z3i) + coh.get_z3_constraints(z3i))

    def generate_linear_constraints(self, ls: LinearSolver):
        opt = Opt(self.af)
        coh = Coh(self.af)
        self.constraints_linear.extend(opt.get_linear_constraints(ls))
        self.constraints_linear.extend(coh.get_linear_constraints(ls))


######################
# semantics by Baier
######################

class CF(Semantics):
    """
    Conflict-freeness semantics: P(A,B) = 0 for all attacks A -> B.
    """
    def generate_constraints(self, si: SolverInstance):
        for edge in self.af.get_edges():
            assignment = tuple(sorted([(edge.node_from, True), (edge.node_to, True)]))
            # noinspection PyTypeChecker
            symbol = si.get_symbol(assignment)
            self.add_constraint(Equals(symbol, Real(0)), edge=edge)

    def generate_z3_constraints(self, z3i: Z3Instance):
        for edge in self.af.get_edges():
            assignment = tuple(sorted([(edge.node_from, True), (edge.node_to, True)]))
            # noinspection PyTypeChecker
            prob_var = z3i.get_prob_var(assignment)
            self.add_z3_constraint(prob_var == 0, edge=edge)

    def generate_linear_constraints(self, ls: LinearSolver):
        for edge in self.af.get_edges():
            assignment = tuple(sorted([(edge.node_from, True), (edge.node_to, True)]))
            # noinspection PyTypeChecker
            full_assignments = asg.sum_out(assignment, self.all_nodes)
            self.constraints_linear.extend(ls.constraints_assignments_equal_value(full_assignments, 0))


# Admissibility

def almost_sure_defense_constraint(si: SolverInstance, node: Node):
    """
    Generate a constraint for an argument A to be almost-surely defended.
    This means for all attackers B->A, the probability that B is in turn attacked (and thus A is defended) is 1.
    That is, P(C1 or C2 or ... or Cn) = 1 for attackers C1...Cn of B, or, equivalently, P(nC1, nC2, ..., nCn) = 0.
    If an attacker B of A has no attackers, then A cannot be almost-surely defended, thus the impossible constraint
    'False' is added.
    """
    parents = list(node.get_parents())
    if len(parents) == 0:
        # unattacked nodes are necessarily defended
        return Bool(True)

    constraints = []
    for attacker in parents:
        attacker_parents: Set[Node] = attacker.get_parents()
        if attacker_parents:
            assignment_list = []
            for defender in attacker_parents:
                assignment_list.append((defender, False))
            assignment = tuple(sorted(assignment_list))
            symbol = si.get_symbol(assignment)
            constraints.append(Equals(symbol, Real(0)))
        else:
            return Bool(False)
    return And(constraints)


def almost_sure_defense_constraint_z3(z3i: Z3Instance, node: Node):
    """
    Generate a constraint for an argument A to be almost-surely defended.
    This means for all attackers B->A, the probability that B is in turn attacked (and thus A is defended) is 1.
    That is, P(C1 or C2 or ... or Cn) = 1 for attackers C1...Cn of B, or, equivalently, P(nC1, nC2, ..., nCn) = 0.
    If an attacker B of A has no attackers, then A cannot be almost-surely defended, thus the impossible constraint
    'False' is added.
    """
    parents = list(node.get_parents())
    if len(parents) == 0:
        # unattacked nodes are necessarily defended
        return z3.BoolVal(True, ctx=z3i.context)

    # special treatment for the first attacker as starting point for nested Ands
    first_attacker = parents[0]
    first_attacker_parents: Set[Node] = first_attacker.get_parents()
    if first_attacker_parents:
        assignment_list = []
        for defender in first_attacker_parents:
            assignment_list.append((defender, False))
        assignment = tuple(sorted(assignment_list))
        prob_var = z3i.get_prob_var(assignment)
        constraint = prob_var == 0
    else:
        return z3.BoolVal(False, ctx=z3i.context)

    for attacker in parents[1:]:
        attacker_parents: Set[Node] = attacker.get_parents()
        if attacker_parents:
            assignment_list = []
            for defender in attacker_parents:
                assignment_list.append((defender, False))
            assignment = tuple(sorted(assignment_list))
            prob_var = z3i.get_prob_var(assignment)
            constraint = z3.And(constraint, prob_var == 0)
        else:
            return z3.BoolVal(False, ctx=z3i.context)
    return constraint


class WAdm(Semantics):
    """
    Weak admissibility semantics: CF and P(A) = 1 implies A is almost-surely defended.
    """
    def generate_constraints(self, si: SolverInstance):
        sCF = CF(self.af)
        self.constraints.extend(sCF.get_constraints(si))
        for node in self.all_nodes:
            if node.is_initial():
                # initial nodes are always almost-surely defended, thus the constraint would become 'p_A == 1 -> True'
                continue
            node_symbol = si.get_node_symbol(node)
            as_defense_constraint = almost_sure_defense_constraint(si, node)
            self.add_constraint(Implies(Equals(node_symbol, Real(1)), as_defense_constraint), node=node)

    def generate_z3_constraints(self, z3i: Z3Instance):
        sCF = CF(self.af)
        self.constraints_z3.extend(sCF.get_z3_constraints(z3i))
        for node in self.all_nodes:
            if node.is_initial():
                # initial nodes are always almost-surely defended, thus the constraint would become 'p_A == 1 -> True'
                continue
            node_prob = z3i.get_node_prob_var(node)
            as_defense_constraint = almost_sure_defense_constraint_z3(z3i, node)
            constraint = z3.Implies(node_prob == 1, as_defense_constraint, ctx=z3i.context)
            self.add_z3_constraint(constraint, node=node)


class PrAdm(Semantics):
    """
    Probabilistic admissibility semantics: CF and for all arguments B with attackers A1...An and attackees C1...Cm,
    P(C1 or ... or Cm) <= P(A1 or ... or An) holds. Equivalently, by negating each side and simplifying, we get
    P(nC1, ..., nCm) >= P(nA1, ..., nAn).
    """
    def generate_constraints(self, si: SolverInstance):
        sCF = CF(self.af)
        self.constraints.extend(sCF.get_constraints(si))
        for node in self.all_nodes:
            node_parents: Set[Node] = node.get_parents()
            if node_parents:
                attacker_assignment = tuple([(attacker, False) for attacker in node_parents])
                attacker_symbol = si.get_symbol(attacker_assignment)
            else:
                # there are no attackers, so the prob of any attacker holding is 0
                # thus, in the negated view, the prob is 1
                attacker_symbol = Real(1)
            node_children: Set[Node] = node.get_children()
            if node_children:
                attackee_assignment = tuple([(attackee, False) for attackee in node_children])
                attackee_symbol = si.get_symbol(attackee_assignment)
            else:
                # there are no attackees, so the prob of any attackee holding is 0
                # thus, in the negated view, the prob is 1
                attackee_symbol = Real(1)
            if node_parents or node_children:
                self.add_constraint(GE(attackee_symbol, attacker_symbol), node=node)

    def generate_z3_constraints(self, z3i: Z3Instance):
        sCF = CF(self.af)
        self.constraints_z3.extend(sCF.get_z3_constraints(z3i))
        for node in self.all_nodes:
            node_parents: Set[Node] = node.get_parents()
            if node_parents:
                attacker_assignment = tuple([(attacker, False) for attacker in node_parents])
                attacker_prob_var = z3i.get_prob_var(attacker_assignment)
            else:
                # there are no attackers, so the prob of any attacker holding is 0
                # thus, in the negated view, the prob is 1
                attacker_prob_var = 1
            node_children: Set[Node] = node.get_children()
            if node_children:
                attackee_assignment = tuple([(attackee, False) for attackee in node_children])
                attackee_prob_var = z3i.get_prob_var(attackee_assignment)
            else:
                # there are no attackees, so the prob of any attackee holding is 0
                # thus, in the negated view, the prob is 1
                attackee_prob_var = 1
            if node_parents or node_children:
                self.add_z3_constraint(attackee_prob_var >= attacker_prob_var, node=node)

    def generate_linear_constraints(self, ls: LinearSolver):
        sCF = CF(self.af)
        self.constraints_linear.extend(sCF.get_linear_constraints(ls))
        attacker_full_assignments = None
        attacker_value = None
        for node in self.all_nodes:
            node_parents: Set[Node] = node.get_parents()
            if node_parents:
                attacker_assignment = tuple([(attacker, False) for attacker in node_parents])
                attacker_full_assignments = asg.sum_out(attacker_assignment, self.all_nodes)
            else:
                # there are no attackers, so the prob of any attacker holding is 0
                # thus, in the negated view, the prob is 1
                attacker_value = 1
            node_children: Set[Node] = node.get_children()
            if node_children:
                attackee_assignment = tuple([(attackee, False) for attackee in node_children])
                attackee_full_assignments = asg.sum_out(attackee_assignment, self.all_nodes)
                if node_parents:
                    c = ls.constraints_assignments_greater_or_equal_assignments(attackee_full_assignments,
                                                                                attacker_full_assignments)
                    self.constraints_linear.extend(c)
                else:
                    c = ls.constraints_assignments_greater_or_equal_value(attackee_full_assignments, attacker_value)
                    self.constraints_linear.extend(c)
            else:
                # there are no attackees, so the prob of any attackee holding is 0
                # thus, in the negated view, the prob is 1
                attackee_value = 1
                if node_parents:
                    c = ls.constraints_assignments_less_or_equal_value(attacker_full_assignments, attackee_value)
                    self.constraints_linear.extend(c)
                else:
                    # constraint would be 1 >= 1
                    pass


class MinAdm(Semantics):
    """
    Min-admissibility semantics: CF and for every argument C, P(C) <= min_{B in Pre(C)} P(OR Pre(B)) holds.
    Equivalently, for all B in Pre(C) with Pre(B) = {A1, ..., An}, it holds P(C) <= 1 - P(nA1, ..., nAn).
    """
    def generate_constraints(self, si: SolverInstance):
        sCF = CF(self.af)
        self.constraints.extend(sCF.get_constraints(si))
        for node in self.all_nodes:
            node_parents = node.get_parents()
            if node_parents:
                node_symbol = si.get_node_symbol(node)
                for attacker in node_parents:
                    attacker_parents: Set[Node] = attacker.get_parents()
                    if attacker_parents:
                        assignment = tuple([(defender, False) for defender in attacker_parents])
                        symbol = si.get_symbol(assignment)
                        self.add_constraint(LE(node_symbol, Minus(Real(1), symbol)), node=node)
                    else:
                        self.add_constraint(Equals(node_symbol, Real(0)), node=node)
            else:
                pass  # minAdm enforces no constraint for initial arguments

    def generate_z3_constraints(self, z3i: Z3Instance):
        sCF = CF(self.af)
        self.constraints_z3.extend(sCF.get_z3_constraints(z3i))
        for node in self.all_nodes:
            node_parents = node.get_parents()
            if node_parents:
                node_prob = z3i.get_node_prob_var(node)
                for attacker in node_parents:
                    attacker_parents: Set[Node] = attacker.get_parents()
                    if attacker_parents:
                        assignment = tuple([(defender, False) for defender in attacker_parents])
                        prob_var = z3i.get_prob_var(assignment)
                        self.add_z3_constraint(node_prob <= 1 - prob_var, node=node)
                    else:
                        self.add_z3_constraint(node_prob == 0, node=node)
            else:
                pass  # minAdm enforces no constraint for initial arguments

    def generate_linear_constraints(self, ls: LinearSolver):
        sCF = CF(self.af)
        self.constraints_linear.extend(sCF.get_linear_constraints(ls))
        for node in self.all_nodes:
            node_parents = node.get_parents()
            if node_parents:
                node_full_assignments = asg.get_node_assignments(node, self.all_nodes)
                for attacker in node_parents:
                    attacker_parents: Set[Node] = attacker.get_parents()
                    if attacker_parents:
                        assignment = tuple([(defender, False) for defender in attacker_parents])
                        full_assignments = asg.sum_out(assignment, self.all_nodes)
                        row_1 = ls.assignments_to_row(node_full_assignments)
                        row_2 = ls.assignments_to_row(full_assignments)
                        # p_node <= 1 - p_attacker_parents
                        # p_node + p_attacker_parents <= 1
                        self.constraints_linear.append((row_1 + row_2, 1))
                    else:
                        c = ls.constraints_assignments_equal_value(node_full_assignments, 0)
                        self.constraints_linear.extend(c)
            else:
                pass  # minAdm enforces no constraint for initial arguments


class JntAdm(Semantics):
    """
    Joint-admissibility semantics: CF and for every argument C, P(C) <= P(AND_{B in Pre(C)} (OR Pre(B))) holds.
    """
    def generate_constraints(self, si: SolverInstance):
        sCF = CF(self.af)
        self.constraints.extend(sCF.get_constraints(si))
        for node in self.all_nodes:
            if node.is_initial():
                # for nodes without attackers, the constraint becomes P(C) <= 1, so we skip it
                continue
            node_symbol = si.get_node_symbol(node)
            # the positive cnf corresponding to AND_{B in Pre(C)} (OR Pre(B))
            cnf = [list(attacker.get_parents()) for attacker in node.get_parents()]
            if [] in cnf:
                # there is an attacker B which is not in turn attacked, thus P(AND_{B in Pre(C)} (OR Pre(B))) = 0
                self.add_constraint(Equals(node_symbol, Real(0)), node=node)
                continue
            cnf_assignments = asg.positive_cnf_assignments(cnf, self.all_nodes)
            symbols = [si.get_symbol(a) for a in cnf_assignments]
            self.add_constraint(LE(node_symbol, Plus(symbols)), node=node)

    def generate_z3_constraints(self, z3i: Z3Instance):
        sCF = CF(self.af)
        self.constraints_z3.extend(sCF.get_z3_constraints(z3i))
        for node in self.all_nodes:
            if node.is_initial():
                # for nodes without attackers, the constraint becomes P(C) <= 1, so we skip it
                continue
            node_prob = z3i.get_node_prob_var(node)
            # the positive cnf corresponding to AND_{B in Pre(C)} (OR Pre(B))
            cnf = [list(attacker.get_parents()) for attacker in node.get_parents()]
            if [] in cnf:
                # there is an attacker B which is not in turn attacked, thus P(AND_{B in Pre(C)} (OR Pre(B))) = 0
                self.add_z3_constraint(node_prob == 0, node=node)
                continue
            cnf_assignments = asg.positive_cnf_assignments(cnf, self.all_nodes)
            prob_vars = [z3i.get_prob_var(a) for a in cnf_assignments]
            # self.add_z3_constraint(node_prob <= z3.Sum(prob_vars), node=node)
            self.add_z3_constraint(node_prob.__le__(z3.Sum(prob_vars)), node=node)

    def generate_linear_constraints(self, ls: LinearSolver):
        sCF = CF(self.af)
        self.constraints_linear.extend(sCF.get_linear_constraints(ls))
        for node in self.all_nodes:
            if node.is_initial():
                # for nodes without attackers, the constraint becomes P(C) <= 1, so we skip it
                continue
            # the positive cnf corresponding to AND_{B in Pre(C)} (OR Pre(B))
            cnf = [list(attacker.get_parents()) for attacker in node.get_parents()]
            if [] in cnf:
                # there is an attacker B which is not in turn attacked, thus P(AND_{B in Pre(C)} (OR Pre(B))) = 0
                self.constraints_linear.extend(ls.constraints_node_equals_value(node, 0))
                continue
            node_assignments = asg.get_node_assignments(node, self.all_nodes)
            cnf_assignments = asg.positive_cnf_assignments(cnf, self.all_nodes)
            c = ls.constraints_assignments_less_or_equal_assignments(node_assignments, cnf_assignments)
            self.constraints_linear.extend(c)


# Completeness

class WCmp(Semantics):
    """
    Weak completeness semantics: CF and P(A) = 1 if and only if A is almost-surely defended.
    """
    def generate_constraints(self, si: SolverInstance):
        sCF = CF(self.af)
        self.constraints.extend(sCF.get_constraints(si))
        for node in self.all_nodes:
            node_symbol = si.get_node_symbol(node)
            if node.is_initial():
                # every initial node is almost-surely defended
                self.add_constraint(Equals(node_symbol, Real(1)), node=node)
                continue
            as_defense_constraint = almost_sure_defense_constraint(si, node)
            self.add_constraint(Iff(Equals(node_symbol, Real(1)), as_defense_constraint), node=node)

    def generate_z3_constraints(self, z3i: Z3Instance):
        sCF = CF(self.af)
        self.constraints_z3.extend(sCF.get_z3_constraints(z3i))
        for node in self.all_nodes:
            node_prob = z3i.get_node_prob_var(node)
            if node.is_initial():
                # every initial node is almost-surely defended
                self.add_z3_constraint(node_prob == 1, node=node)
                continue
            as_defense_constraint = almost_sure_defense_constraint_z3(z3i, node)
            self.add_z3_constraint(z3.Implies(node_prob == 1, as_defense_constraint, ctx=z3i.context), node=node)
            self.add_z3_constraint(z3.Implies(as_defense_constraint, node_prob == 1, ctx=z3i.context), node=node)


def add_prCmp_constraints(sem: Semantics, nodes: List[Node], si: SolverInstance):
    """
    Given a list of arguments, returns the constraints P(C) >= P(AND_{B in Pre(C)} (OR Pre(B))) for every argument C.
    """
    for node in nodes:
        node_symbol = si.get_node_symbol(node)
        if node.is_initial():
            # for nodes without attackers, the constraint becomes P(C) >= 1
            sem.add_constraint(Equals(node_symbol, Real(1)), node=node)
            continue
        # the positive cnf corresponding to AND_{B in Pre(C)} (OR Pre(B))
        cnf = [list(attacker.get_parents()) for attacker in node.get_parents()]
        if [] in cnf:
            # there is an attacker B which is not in turn attacked, thus P(AND_{B in Pre(C)} (OR Pre(B))) = 0
            # the constraint would be P(C) >= 0, so we skip it
            continue
        cnf_assignments = asg.positive_cnf_assignments(cnf, nodes)
        cnf_symbols = [si.get_symbol(a) for a in cnf_assignments]
        sem.add_constraint(GE(node_symbol, Plus(cnf_symbols)), node=node)


def add_prCmp_z3_constraints(sem: Semantics, nodes: List[Node], z3i: Z3Instance):
    """
    Given a list of arguments, returns the constraints P(C) >= P(AND_{B in Pre(C)} (OR Pre(B))) for every argument C.
    """
    for node in nodes:
        node_prob = z3i.get_node_prob_var(node)
        if node.is_initial():
            # for nodes without attackers, the constraint becomes P(C) >= 1
            sem.add_z3_constraint(node_prob == 1, node=node)
            continue
        # the positive cnf corresponding to AND_{B in Pre(C)} (OR Pre(B))
        cnf = [list(attacker.get_parents()) for attacker in node.get_parents()]
        if [] in cnf:
            # there is an attacker B which is not in turn attacked, thus P(AND_{B in Pre(C)} (OR Pre(B))) = 0
            # the constraint would be P(C) >= 0, so we skip it
            continue
        cnf_assignments = asg.positive_cnf_assignments(cnf, nodes)
        prob_vars = [z3i.get_prob_var(a) for a in cnf_assignments]
        sem.add_z3_constraint(node_prob >= z3.Sum(prob_vars), node=node)


def prCmp_linear_constraints(nodes: List[Node], ls: LinearSolver):
    """
    Given a list of arguments, returns the constraints P(C) >= P(AND_{B in Pre(C)} (OR Pre(B))) for every argument C.
    """
    constraints = []
    for node in nodes:
        node_assignments = asg.get_node_assignments(node, ls.af.get_nodes())
        if node.is_initial():
            # for nodes without attackers, the constraint becomes P(C) >= 1, and thus P(C) == 1
            constraints.extend(ls.constraints_assignments_equal_value(node_assignments, 1))
            continue
        # the positive cnf corresponding to AND_{B in Pre(C)} (OR Pre(B))
        cnf = [list(attacker.get_parents()) for attacker in node.get_parents()]
        if [] in cnf:
            # there is an attacker B which is not in turn attacked, thus P(AND_{B in Pre(C)} (OR Pre(B))) = 0
            # the constraint would be P(C) >= 0, so we skip it
            continue
        cnf_assignments = asg.positive_cnf_assignments(cnf, nodes)
        c = ls.constraints_assignments_greater_or_equal_assignments(node_assignments, cnf_assignments)
        constraints.extend(c)
    return constraints


class PrCmp(Semantics):
    """
    Probabilistic completeness semantics: Prob. admissible and P(C) >= P(AND_{B in Pre(C)} (OR Pre(B))) holds for all
    arguments C.
    """
    def generate_constraints(self, si: SolverInstance):
        prAdm = PrAdm(self.af)
        self.constraints.extend(prAdm.get_constraints(si))
        add_prCmp_constraints(self, self.all_nodes, si)

    def generate_z3_constraints(self, z3i: Z3Instance):
        prAdm = PrAdm(self.af)
        self.constraints_z3.extend(prAdm.get_z3_constraints(z3i))
        add_prCmp_z3_constraints(self, self.all_nodes, z3i)

    def generate_linear_constraints(self, ls: LinearSolver):
        prAdm = PrAdm(self.af)
        self.constraints_linear.extend(prAdm.get_linear_constraints(ls))
        self.constraints_linear.extend(prCmp_linear_constraints(self.all_nodes, ls))


class MinCmp(Semantics):
    """
    Min-complete semantics: Min-admissible and P(C) >= P(AND_{B in Pre(C)} (OR Pre(B))) holds for all arguments C.
    """
    def generate_constraints(self, si: SolverInstance):
        minAdm = MinAdm(self.af)
        self.constraints.extend(minAdm.get_constraints(si))
        add_prCmp_constraints(self, self.all_nodes, si)

    def generate_z3_constraints(self, z3i: Z3Instance):
        minAdm = MinAdm(self.af)
        self.constraints_z3.extend(minAdm.get_z3_constraints(z3i))
        add_prCmp_z3_constraints(self, self.all_nodes, z3i)

    def generate_linear_constraints(self, ls: LinearSolver):
        minAdm = MinAdm(self.af)
        self.constraints_linear.extend(minAdm.get_linear_constraints(ls))
        self.constraints_linear.extend(prCmp_linear_constraints(self.all_nodes, ls))


class JntCmp(Semantics):
    """
    Joint-complete semantics: CF and P(C) = P(AND_{B in Pre(C)} (OR Pre(B))) holds for all arguments C.
    """
    def generate_constraints(self, si: SolverInstance):
        sCF = CF(self.af)
        self.constraints.extend(sCF.get_constraints(si))
        for node in self.all_nodes:
            node_symbol = si.get_node_symbol(node)
            if node.is_initial():
                # for nodes without attackers, the constraint becomes P(C) == 1
                self.add_constraint(Equals(node_symbol, Real(1)), node=node)
                continue
            # the positive cnf corresponding to AND_{B in Pre(C)} (OR Pre(B))
            cnf = [list(attacker.get_parents()) for attacker in node.get_parents()]
            if [] in cnf:
                # there is an attacker B which is not in turn attacked, thus P(AND_{B in Pre(C)} (OR Pre(B))) = 0
                self.add_constraint(Equals(node_symbol, Real(0)), node=node)
                continue
            cnf_assignments = asg.positive_cnf_assignments(cnf, self.all_nodes)
            cnf_symbols = [si.get_symbol(a) for a in cnf_assignments]
            self.add_constraint(Equals(node_symbol, Plus(cnf_symbols)), node=node)

    def generate_z3_constraints(self, z3i: Z3Instance):
        sCF = CF(self.af)
        self.constraints_z3.extend(sCF.get_z3_constraints(z3i))
        for node in self.all_nodes:
            node_prob = z3i.get_node_prob_var(node)
            if node.is_initial():
                # for nodes without attackers, the constraint becomes P(C) == 1
                self.add_z3_constraint(node_prob == 1, node=node)
                continue
            # the positive cnf corresponding to AND_{B in Pre(C)} (OR Pre(B))
            cnf = [list(attacker.get_parents()) for attacker in node.get_parents()]
            if [] in cnf:
                # there is an attacker B which is not in turn attacked, thus P(AND_{B in Pre(C)} (OR Pre(B))) = 0
                self.add_z3_constraint(node_prob == 0, node=node)
                continue
            cnf_assignments = asg.positive_cnf_assignments(cnf, self.all_nodes)
            prob_vars = [z3i.get_prob_var(a) for a in cnf_assignments]
            # self.add_z3_constraint(node_prob == z3.Sum(prob_vars), node=node)
            self.add_z3_constraint(node_prob.__eq__(z3.Sum(prob_vars)), node=node)

    def generate_linear_constraints(self, ls: LinearSolver):
        sCF = CF(self.af)
        self.constraints_linear.extend(sCF.get_linear_constraints(ls))
        for node in self.all_nodes:
            if node.is_initial():
                # for nodes without attackers, the constraint becomes P(C) == 1
                c = ls.constraints_node_equals_value(node, 1)
                self.constraints_linear.extend(c)
                continue
            # the positive cnf corresponding to AND_{B in Pre(C)} (OR Pre(B))
            cnf = [list(attacker.get_parents()) for attacker in node.get_parents()]
            if [] in cnf:
                # there is an attacker B which is not in turn attacked, thus P(AND_{B in Pre(C)} (OR Pre(B))) = 0
                c = ls.constraints_node_equals_value(node, 0)
                self.constraints_linear.extend(c)
                continue
            node_assignments = asg.get_node_assignments(node, self.all_nodes)
            cnf_assignments = asg.positive_cnf_assignments(cnf, self.all_nodes)
            c = ls.constraints_assignments_equal_assignments(node_assignments, cnf_assignments)
            self.constraints_linear.extend(c)


######################################################
# Element-wise lifted versions of classical semantics
######################################################

class ElmClassicalSemantics(Semantics):
    """
    Element-wise lifted classical semantics: If a probabilistic variable from the full joint distribution has positive
    probability, it needs to correspond to an extension of the classical semantics. That is, all prob vars
    corresponding to other extensions need to have probability zero.
    """
    def __init__(self, classical_semantics: Type[classical.ClassicalSemantics], af: DPGM, complemented=False):
        self.classical_semantics = classical_semantics
        super().__init__(af, complemented)

    def generate_constraints(self, si: SolverInstance):
        sem = self.classical_semantics(self.af)
        sem_extensions = sem.get_extensions()
        constraint_list = []
        for assignment in si.full_assignments:
            if asg.to_extension(assignment) not in sem_extensions:
                symbol = si.get_symbol(assignment)
                constraint_list.append(Equals(symbol, Real(0)))
        self.add_constraint(And(constraint_list))

    def generate_z3_constraints(self, z3i: Z3Instance):
        sem = self.classical_semantics(self.af)
        sem_extensions = sem.get_extensions()
        constraint_list = []
        for assignment in z3i.full_joint_assignments:
            if asg.to_extension(assignment) not in sem_extensions:
                prob_var = z3i.get_prob_var(assignment)
                constraint_list.append(prob_var == 0)
        self.add_z3_constraint(z3.And(constraint_list))

    def generate_linear_constraints(self, ls: LinearSolver):
        sem = self.classical_semantics(self.af)
        sem_extensions = sem.get_extensions()
        non_adm_assignments = []
        for assignment in asg.generate(self.all_nodes):
            if asg.to_extension(assignment) not in sem_extensions:
                non_adm_assignments.append(assignment)
        self.constraints_linear.extend(ls.constraints_assignments_equal_value(non_adm_assignments, 0))


class ElmCF(ElmClassicalSemantics):
    """
    Element-wise conflict-free semantics.
    """
    def __init__(self, af: DPGM, complemented=False):
        super().__init__(classical.CF, af, complemented)


class ElmAdm(ElmClassicalSemantics):
    """
    Element-wise admissible semantics.
    """
    def __init__(self, af: DPGM, complemented=False):
        super().__init__(classical.Admissible, af, complemented)


class ElmCmp(ElmClassicalSemantics):
    """
    Element-wise complete semantics.
    """
    def __init__(self, af: DPGM, complemented=False):
        super().__init__(classical.Complete, af, complemented)


class ElmGrn(ElmClassicalSemantics):
    """
    Element-wise grounded semantics.
    """
    def __init__(self, af: DPGM, complemented=False):
        super().__init__(classical.Grounded, af, complemented)


class ElmPrf(ElmClassicalSemantics):
    """
    Element-wise preferred semantics.
    """
    def __init__(self, af: DPGM, complemented=False):
        super().__init__(classical.Preferred, af, complemented)


class ElmSStl(ElmClassicalSemantics):
    """
    Element-wise semi-stable semantics.
    """
    def __init__(self, af: DPGM, complemented=False):
        super().__init__(classical.SemiStable, af, complemented)


class ElmStl(ElmClassicalSemantics):
    """
    Element-wise stable semantics.
    """
    def __init__(self, af: DPGM, complemented=False):
        super().__init__(classical.Stable, af, complemented)


################################################
# Experimental semantics generalizing labelings
################################################

# class LabAdm(Semantics):
#     def generate_z3_constraints(self, z3i: Z3Instance):
#         sCF = CF(self.af)
#         self.constraints_z3.extend(sCF.get_z3_constraints(z3i))
#         for node in self.all_nodes:
#             node_prob = z3i.get_node_prob_var(node)
#             or_list = []
#             for parent_node in node.get_parents():
#                 parent_node_prob = z3i.get_node_prob_var(parent_node)
#                 or_list.append(parent_node_prob == 1)
#             self.add_z3_constraint(z3.Implies(node_prob == 0, z3.Or(or_list, z3i.context), z3i.context), node=node)
#
#
# class SLabAdm(Semantics):
#     def generate_z3_constraints(self, z3i: Z3Instance):
#         sCF = CF(self.af)
#         self.constraints_z3.extend(sCF.get_z3_constraints(z3i))
#         for node in self.all_nodes:
#             node_prob = z3i.get_node_prob_var(node)
#             if node.is_initial():
#                 continue
#             or_list = []
#             for parent_node in node.get_parents():
#                 parent_node_prob = z3i.get_node_prob_var(parent_node)
#                 or_list.append(parent_node_prob == 1)
#             self.add_z3_constraint(z3.Implies(node_prob == 0, z3.Or(or_list, z3i.context), z3i.context), node=node)
#
#
# def open_interval_constraint(prob_var, context, low=0, high=1):
#     return z3.And(prob_var > low, prob_var < high, context)
#
#
# class LabCmp(Semantics):
#     def generate_z3_constraints(self, z3i: Z3Instance):
#         labAdm = LabAdm(self.af)
#         self.constraints_z3.extend(labAdm.get_z3_constraints(z3i))
#         for node in self.all_nodes:
#             node_prob = z3i.get_node_prob_var(node)
#             node_undec = open_interval_constraint(node_prob, z3i.context)
#             or_list = []
#             for parent_node in node.get_parents():
#                 parent_node_prob = z3i.get_node_prob_var(parent_node)
#                 self.add_z3_constraint(z3.Implies(node_undec, parent_node_prob < 1, z3i.context), node=node)
#                 parent_undec = open_interval_constraint(parent_node_prob, z3i.context)
#                 or_list.append(parent_undec)
#             self.add_z3_constraint(z3.Implies(node_undec, z3.Or(or_list, z3i.context), z3i.context), node=node)
#
#
# class SLabCmp(Semantics):
#     def generate_z3_constraints(self, z3i: Z3Instance):
#         sLabAdm = SLabAdm(self.af)
#         self.constraints_z3.extend(sLabAdm.get_z3_constraints(z3i))
#         for node in self.all_nodes:
#             node_prob = z3i.get_node_prob_var(node)
#             if node.is_initial():
#                 continue
#             node_undec = open_interval_constraint(node_prob, z3i.context)
#             or_list = []
#             for parent_node in node.get_parents():
#                 parent_node_prob = z3i.get_node_prob_var(parent_node)
#                 self.add_z3_constraint(z3.Implies(node_undec, parent_node_prob < 1, z3i.context), node=node)
#                 parent_undec = open_interval_constraint(parent_node_prob, z3i.context)
#                 or_list.append(parent_undec)
#             self.add_z3_constraint(z3.Implies(node_undec, z3.Or(or_list, z3i.context), z3i.context), node=node)


#####################
# semantics by Käfer
#####################

class WNorS(Semantics):
    """
    Weak not-or semantics without CF: Generates constraints P(A) <= P(-B1,-B2,...,-Bi) for every non-initial argument A
    with attackers B1, ..., Bi for some i > 0.
    """
    def generate_constraints(self, si: SolverInstance):
        for node in self.all_nodes:
            if node.is_initial():
                continue  # for initial nodes, the constraint would be 'p_A <= 1' which already exists
            node_symbol = si.get_node_symbol(node)
            assignment = tuple([(attacker, False) for attacker in node.get_parents()])
            # noinspection PyTypeChecker
            attackers_symbol = si.get_symbol(assignment)
            self.add_constraint(LE(node_symbol, attackers_symbol), node=node)

    def generate_z3_constraints(self, z3i: Z3Instance):
        for node in self.all_nodes:
            if node.is_initial():
                continue  # for initial nodes, the constraint would be 'p_A <= 1' which already exists
            node_prob = z3i.get_node_prob_var(node)
            assignment = tuple([(attacker, False) for attacker in node.get_parents()])
            # noinspection PyTypeChecker
            attackers_prob = z3i.get_prob_var(assignment)
            # self.add_z3_constraint(node_prob <= attackers_prob, node=node)
            self.add_z3_constraint(node_prob.__le__(attackers_prob), node=node)

    def generate_linear_constraints(self, ls: LinearSolver):
        raise NotImplementedError  # TODO


class NorS(Semantics):
    """
    Not-or semantics without CF: Generates constraints P(A) = P(-B1,-B2,...,-Bi) for every non-initial argument A with
    attackers B1, ..., Bi for some i > 0.
    """
    def generate_constraints(self, si: SolverInstance):
        for node in self.all_nodes:
            if node.is_initial():
                continue  # for initial nodes, the constraint would be 'p_A <= 1' which already exists
            node_symbol = si.get_node_symbol(node)
            assignment = tuple([(attacker, False) for attacker in node.get_parents()])
            # noinspection PyTypeChecker
            attackers_symbol = si.get_symbol(assignment)
            self.add_constraint(Equals(node_symbol, attackers_symbol), node=node)

    def generate_z3_constraints(self, z3i: Z3Instance):
        for node in self.all_nodes:
            if node.is_initial():
                continue
            node_prob = z3i.get_node_prob_var(node)
            assignment = tuple([(attacker, False) for attacker in node.get_parents()])
            # noinspection PyTypeChecker
            attackers_prob = z3i.get_prob_var(assignment)
            # self.add_z3_constraint(node_prob == attackers_prob, node=node)
            self.add_z3_constraint(node_prob.__eq__(attackers_prob), node=node)

    def generate_linear_constraints(self, ls: LinearSolver):
        raise NotImplementedError  # TODO


class SNorS(Semantics):
    """
    Strong not-or semantics without CF. Generates constraints P(A) = 1 for initial arguments and
    P(A) = P(-B1,-B2,...,-Bi) for every non-initial argument A with attackers B1, ..., Bi for some i>0.
    Also the intersection of Nor and Fou.
    """
    def generate_constraints(self, si: SolverInstance):
        for node in self.all_nodes:
            node_symbol = si.get_node_symbol(node)
            if node.is_initial():
                self.add_constraint(Equals(node_symbol, Real(1)), node=node)
            else:
                assignment = tuple([(attacker, False) for attacker in node.get_parents()])
                # noinspection PyTypeChecker
                attackers_symbol = si.get_symbol(assignment)
                self.add_constraint(Equals(node_symbol, attackers_symbol), node=node)

    def generate_z3_constraints(self, z3i: Z3Instance):
        for node in self.all_nodes:
            node_prob = z3i.get_node_prob_var(node)
            if node.is_initial():
                self.add_z3_constraint(node_prob == 1, node=node)
            else:
                assignment: List[Tuple[Node, bool]] = [(attacker, False) for attacker in node.get_parents()]
                attackers_prob = z3i.get_prob_var(tuple(assignment))
                # self.add_z3_constraint(node_prob == attackers_prob, node=node)
                self.add_z3_constraint(node_prob.__eq__(attackers_prob), node=node)

    def generate_linear_constraints(self, ls: LinearSolver):
        raise NotImplementedError  # TODO


class WNor(Semantics):
    """
    Weak not-or semantics: Generates constraints for conflict-freeness and P(A) <= P(-B1,-B2,...,-Bi) for every
    non-initial argument A with attackers B1, ..., Bi for some i > 0.
    """
    def generate_constraints(self, si: SolverInstance):
        sCF = CF(self.af)
        self.constraints.extend(sCF.get_constraints(si))
        sWNorS = WNorS(self.af)
        self.constraints.extend(sWNorS.get_constraints(si))

    def generate_z3_constraints(self, z3i: Z3Instance):
        sCF = CF(self.af)
        self.constraints_z3.extend(sCF.get_z3_constraints(z3i))
        for node in self.all_nodes:
            if node.is_initial():
                continue  # for initial nodes, the constraint would be 'p_A <= 1' which already exists
            node_prob = z3i.get_node_prob_var(node)
            assignment: List[Tuple[Node, bool]] = [(attacker, False) for attacker in node.get_parents()]
            attackers_prob = z3i.get_prob_var(tuple(assignment))
            # self.add_z3_constraint(node_prob <= attackers_prob, node=node)
            self.add_z3_constraint(node_prob.__le__(attackers_prob), node=node)

    def generate_linear_constraints(self, ls: LinearSolver):
        sCF = CF(self.af)
        self.constraints_linear.extend(sCF.get_linear_constraints(ls))
        sWNorS = WNorS(self.af)
        self.constraints_linear.extend(sWNorS.get_linear_constraints(ls))


class Nor(Semantics):
    """
    Not-or semantics: Generates constraints for conflict-freeness and P(A) = P(-B1,-B2,...,-Bi) for every non-initial
    argument A with attackers B1, ..., Bi for some i > 0.
    """
    def generate_constraints(self, si: SolverInstance):
        sCF = CF(self.af)
        self.constraints.extend(sCF.get_constraints(si))
        sNorS = NorS(self.af)
        self.constraints.extend(sNorS.get_constraints(si))

    def generate_z3_constraints(self, z3i: Z3Instance):
        sCF = CF(self.af)
        self.constraints_z3.extend(sCF.get_z3_constraints(z3i))
        for node in self.all_nodes:
            if node.is_initial():
                continue
            node_prob = z3i.get_node_prob_var(node)
            assignment: List[Tuple[Node, bool]] = [(attacker, False) for attacker in node.get_parents()]
            attackers_prob = z3i.get_prob_var(tuple(assignment))
            # self.add_z3_constraint(node_prob == attackers_prob, node=node)
            self.add_z3_constraint(node_prob.__eq__(attackers_prob), node=node)

    def generate_linear_constraints(self, ls: LinearSolver):
        sCF = CF(self.af)
        self.constraints_linear.extend(sCF.get_linear_constraints(ls))
        sNorS = NorS(self.af)
        self.constraints_linear.extend(sNorS.get_linear_constraints(ls))


class SNor(Semantics):
    """
    Strong not-or semantics: Generates constraints for conflict-freeness, P(A) = 1 for initial arguments and
    P(A) = P(-B1,-B2,...,-Bi) for every non-initial argument A with attackers B1, ..., Bi for some i>0.
    Also the intersection of Nor and Fou and CF.
    """
    def generate_constraints(self, si: SolverInstance):
        sCF = CF(self.af)
        self.constraints.extend(sCF.get_constraints(si))
        sSNorS = SNorS(self.af)
        self.constraints.extend(sSNorS.get_constraints(si))

    def generate_z3_constraints(self, z3i: Z3Instance):
        sCF = CF(self.af)
        self.constraints_z3.extend(sCF.get_z3_constraints(z3i))
        for node in self.all_nodes:
            node_prob = z3i.get_node_prob_var(node)
            if node.is_initial():
                self.add_z3_constraint(node_prob == 1, node=node)
            else:
                assignment: List[Tuple[Node, bool]] = [(attacker, False) for attacker in node.get_parents()]
                attackers_prob = z3i.get_prob_var(tuple(assignment))
                # self.add_z3_constraint(node_prob == attackers_prob, node=node)
                self.add_z3_constraint(node_prob.__eq__(attackers_prob), node=node)

    def generate_linear_constraints(self, ls: LinearSolver):
        sCF = CF(self.af)
        self.constraints_linear.extend(sCF.get_linear_constraints(ls))
        sSNorS = SNorS(self.af)
        self.constraints_linear.extend(sSNorS.get_linear_constraints(ls))


class AF(Semantics):
    """
    Special semantics to add constraints for probability values or intervals specified in the AF input file.
    """
    def generate_constraints(self, si: SolverInstance):
        for node in self.all_nodes:
            node_symbol = si.get_node_symbol(node)
            if node.value is not None:
                assert node.interval is None, "node with both value and interval set"
                value = nice_fraction(node.value)
                self.add_constraint(Equals(node_symbol, Real(value)), node=node)
            elif node.interval is not None:
                low, high = node.interval
                low = nice_fraction(low)
                high = nice_fraction(high)
                self.add_constraint(And(LE(node_symbol, Real(high)), GE(node_symbol, Real(low))), node=node)

    def generate_z3_constraints(self, z3i: Z3Instance):
        for node in self.all_nodes:
            node_prob = z3i.get_node_prob_var(node)
            if node.value is not None:
                assert node.interval is None, "node with both value and interval set"
                # self.add_z3_constraint(node_prob == node.value, node=node)
                self.add_z3_constraint(node_prob.__eq__(node.value), node=node)
            elif node.interval is not None:
                low, high = node.interval
                # self.add_z3_constraint(node_prob >= low, node=node)
                # self.add_z3_constraint(node_prob <= high, node=node)
                self.add_z3_constraint(node_prob.__ge__(low), node=node)
                self.add_z3_constraint(node_prob.__le__(high), node=node)

    def generate_linear_constraints(self, ls: LinearSolver):
        for node in self.all_nodes:
            node_full_assignments = asg.get_node_assignments(node, self.all_nodes)
            # node_prob = z3i.get_node_prob_var(node)
            if node.value is not None:
                assert node.interval is None, "node with both value and interval set"
                c = ls.constraints_assignments_equal_value(node_full_assignments, node.value)
                self.constraints_linear.extend(c)
            elif node.interval is not None:
                low, high = node.interval
                c1 = ls.constraints_assignments_greater_or_equal_value(node_full_assignments, low)
                c2 = ls.constraints_assignments_less_or_equal_value(node_full_assignments, high)
                self.constraints_linear.extend(c1)
                self.constraints_linear.extend(c2)


class NNor(Semantics):
    """
    Noisy not-or semantics:  TODO
    per default without prior probabilities
    p_A_nB_nC = 1 * p_nB_nC
    p_A_nB_C  = (1 - t) * p_nB_C
    p_A_B_nC  = (1 - s) * p_B_nC
    p_A_B_C   = (1 - s ) * (1 - t) * p_B_C
    """
    def __init__(self, af: DPGM, use_prior_probs=False, complemented=False):
        self.use_prior_probs = use_prior_probs
        super().__init__(af, complemented)

    def generate_z3_constraints(self, z3i: Z3Instance):
        # we use attack strength, so a prob_var should be created for each edge
        z3i.generate_edge_vars()

        for node in self.all_nodes:

            node_prior_prob_var = z3i.create_real_var("pr_" + node.name)  # e.g. pr_A
            if self.use_prior_probs:
                if node.value is not None:
                    # self.add_z3_constraint(node_prior_prob_var == node.value, node=node)
                    self.add_z3_constraint(node_prior_prob_var.__eq__(node.value), node=node)
                elif node.interval is not None:
                    i_min, i_max = node.interval
                    constraint = z3.And(node_prior_prob_var >= i_min, node_prior_prob_var <= i_max)
                    self.add_z3_constraint(constraint, node=node)
                else:
                    self.add_z3_constraint(z3.And(node_prior_prob_var >= 0, node_prior_prob_var <= 1), node=node)
            else:
                self.add_z3_constraint(node_prior_prob_var == 1, node=node)

            if node.is_initial():
                node_prob = z3i.get_node_prob_var(node)
                # self.add_z3_constraint(node_prob == node_prior_prob_var, node=node)
                self.add_z3_constraint(node_prob.__eq__(node_prior_prob_var), node=node)
                continue

            for parent_assignment in asg.generate(node.get_parents()):
                parent_prob_var = z3i.get_prob_var(parent_assignment)  # e.g. p_nB_C
                assignment = tuple(sorted(list(parent_assignment) + [(node, True)]))
                prob_var = z3i.get_prob_var(assignment)  # e.g. p_A_nB_C

                product = node_prior_prob_var * parent_prob_var
                for (parent, status) in parent_assignment:
                    if status:
                        edge = node.get_parent_edge(parent)
                        edge_var = z3i.get_edge_var(edge)
                        product *= (1 - edge_var)
                # self.add_z3_constraint(prob_var == product, node=node)
                self.add_z3_constraint(prob_var.__eq__(product), node=node)


class NNorAF(NNor):
    """
    Noisy not-or semantics:  TODO
    with prior probabilities
    p_A_nB_nC = prA * p_nB_nC
    p_A_nB_C  = prA * (1 - t) * p_nB_C
    p_A_B_nC  = prA * (1 - s) * p_B_nC
    p_A_B_C   = prA * (1 - s ) * (1 - t) * p_B_C
    """
    def __init__(self, af: DPGM, complemented=False):
        super().__init__(af, use_prior_probs=True, complemented=complemented)


class CFs(Semantics):
    """
    Conflict-freeness semantics with attack strengths: P(A,B) <= 1-s for all attacks A -s-> B.
    """
    def generate_z3_constraints(self, z3i: Z3Instance):
        # we use attack strength, so a prob_var should be created for each edge
        z3i.generate_edge_vars()

        for edge in self.af.get_edges():
            edge_var = z3i.get_edge_var(edge)
            assignment: List[Tuple[Node, bool]] = [(edge.node_from, True), (edge.node_to, True)]
            prob_var = z3i.get_prob_var(tuple(assignment))
            self.add_z3_constraint(prob_var <= 1 - edge_var, edge=edge)


class StrengthCF(Semantics):
    """
    Alternative conflict-freeness semantics with attack strengths: "P(A,B) <= (1-s) * P(A)" for all attacks A -s-> B.
    """
    def generate_z3_constraints(self, z3i: Z3Instance):
        # we use attack strength, so a prob_var should be created for each edge
        z3i.generate_edge_vars()

        for edge in self.af.get_edges():
            edge_var = z3i.get_edge_var(edge)
            assignment: List[Tuple[Node, bool]] = [(edge.node_from, True), (edge.node_to, True)]
            prob_var = z3i.get_prob_var(tuple(assignment))
            node_to_prob = z3i.get_node_prob_var(edge.node_to)
            self.add_z3_constraint(prob_var <= (1 - edge_var) * node_to_prob, edge=edge)


########################################
# Strength and Support
########################################

class StrengthSupportCF(Semantics):
    """
    Treats arrows as support with strengths: "P(A,B) >= s * P(A)" for all supports A -s-> B.
    """
    def generate_z3_constraints(self, z3i: Z3Instance):
        # we use attack strength, so a prob_var should be created for each edge
        z3i.generate_edge_vars()

        for edge in self.af.get_edges():
            edge_var = z3i.get_edge_var(edge)
            assignment: List[Tuple[Node, bool]] = [(edge.node_from, True), (edge.node_to, True)]
            prob_var = z3i.get_prob_var(tuple(assignment))
            node_to_prob = z3i.get_node_prob_var(edge.node_to)
            # self.add_z3_constraint(prob_var >= edge_var * node_to_prob, edge=edge)
            self.add_z3_constraint(prob_var.__ge__(edge_var * node_to_prob), edge=edge)


########################################
# Dirac versions of classical semantics
########################################

class DiracClassicalSemantics(Semantics):
    """
    Lifts classical semantics into probabilistic semantics by admitting the Dirac-distributions corresponding to the
    classical semantics' extensions.
    E.g., a distribution P is Dirac-admissible if it is a Dirac distribution and P(beta) = 1 for some assignment beta
    implies that the set of arguments holding in beta is an admissible set.

    This is implemented by adding a constraint "P(beta_1) = 1 OR P(beta_2) = 1 OR ... OR P(beta_n) = 1" where beta_i are
    the assignments corresponding to the respective semantics' extensions. Note that this constraint also ensures that
    distributions satisfying it are Dirac distributions, so no additional Dirac constraints are added.
    """
    def __init__(self, classical_semantics: Type[classical.ClassicalSemantics], af: DPGM, complemented=False):
        self.classical_semantics = classical_semantics
        super().__init__(af, complemented)

    def generate_constraints(self, si: SolverInstance):
        cs = self.classical_semantics(self.af)
        or_list = []
        for extension in cs.get_extensions():
            assignment = asg.from_extension(extension, self.all_nodes)
            symbol = si.get_symbol(assignment)
            or_list.append(Equals(symbol, Real(1)))
        self.add_constraint(Or(or_list))

    def generate_z3_constraints(self, z3i: Z3Instance):
        cs = self.classical_semantics(self.af)
        or_list = []
        for extension in cs.get_extensions():
            assignment = asg.from_extension(extension, self.all_nodes)
            prob_var = z3i.get_prob_var(assignment)
            or_list.append(prob_var == 1)
        if or_list:
            self.add_z3_constraint(z3.Or(or_list))
        else:
            self.add_z3_constraint(z3.BoolVal(False, ctx=z3i.context))


class DiracCF(DiracClassicalSemantics):
    """
    Dirac-conflict-free semantics: Only allows Dirac-distributions corresponding to conflict-free extensions.
    """
    def __init__(self, af: DPGM, complemented=False):
        super().__init__(classical.CF, af, complemented)


class DiracAdm(DiracClassicalSemantics):
    """
    Dirac-admissible semantics: Only allows Dirac-distributions corresponding to admissible extensions.
    """
    def __init__(self, af: DPGM, complemented=False):
        super().__init__(classical.Admissible, af, complemented)


class DiracCmp(DiracClassicalSemantics):
    """
    Dirac-complete semantics: Only allows Dirac-distributions corresponding to complete extensions.
    """
    def __init__(self, af: DPGM, complemented=False):
        super().__init__(classical.Complete, af, complemented)


class DiracGrn(DiracClassicalSemantics):
    """
    Dirac-grounded semantics: Only allows Dirac-distributions corresponding to grounded extensions.
    """
    def __init__(self, af: DPGM, complemented=False):
        super().__init__(classical.Grounded, af, complemented)


class DiracPrf(DiracClassicalSemantics):
    """
    Dirac-preferred semantics: Only allows Dirac-distributions corresponding to preferred extensions.
    """
    def __init__(self, af: DPGM, complemented=False):
        super().__init__(classical.Preferred, af, complemented)


class DiracSStl(DiracClassicalSemantics):
    """
    Dirac-semi-stable semantics: Only allows Dirac-distributions corresponding to semi-stable extensions.
    """
    def __init__(self, af: DPGM, complemented=False):
        super().__init__(classical.SemiStable, af, complemented)


class DiracStl(DiracClassicalSemantics):
    """
    Dirac-stable semantics: Only allows Dirac-distributions corresponding to stable extensions.
    """
    def __init__(self, af: DPGM, complemented=False):
        super().__init__(classical.Stable, af, complemented)


###################
# Helper functions
###################

def get_semantics_class_by_name(name: str):
    """
    By dark magic, get the semantics class given its name as string, e.g. "wNor" to yield the class WNor.

    :param name: the name of the semantics (case insensitive)
    :return: The semantics class corresponding to the name
    """
    queue = Semantics.__subclasses__().copy()
    while queue:
        semantics_class = queue.pop()
        if name.lower() == semantics_class.__name__.lower():
            return semantics_class
        queue.extend(semantics_class.__subclasses__())
    return None


def all_semantics_names():
    """
    :return: A list of the names of all semantics declared in this file
    """
    # helper classes which do not form semantics themselves
    exceptions = ["ElmClassicalSemantics", "DiracClassicalSemantics"]

    names = []
    # use a queue to recursively also get the names of subclasses
    queue = Semantics.__subclasses__().copy()
    while queue:
        semantics_class = queue.pop()
        queue.extend(semantics_class.__subclasses__())
        if semantics_class.__name__ not in exceptions:
            names.append(semantics_class.__name__)
    return names


def nice_fraction(x, max_fraction_denominator=MAX_FRACTION_DENOMINATOR) -> Fraction:
    """

    >>> Fraction(0.7)
    Fraction(3152519739159347, 4503599627370496)
    >>> nice_fraction(0.7)
    Fraction(7, 10)

    """
    return Fraction(x).limit_denominator(max_fraction_denominator)
